/*************************************************************
 postGIS - geometric types for postgres

 This software is copyrighted (2001).
 
 This is free software; you can redistribute it and/or modify
 it under the GNU General Public Licence.  See the file "COPYING".

 More Info?  See the documentation, join the mailing list 
 (postgis@yahoogroups.com), or see the web page
 (http://postgis.refractions.net).

 *************************************************************/



#include "postgres.h"


#include <math.h>
#include <float.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"


#include "fmgr.h"


#include "postgis.h"
#include "utils/elog.h"


#define SHOW_DIGS_DOUBLE 15
#define MAX_DIGS_DOUBLE (SHOW_DIGS_DOUBLE + 6 + 1 + 3 +1)


// #define DEBUG_GIST
//#define DEBUG_GIST2


//prototypes 

     int isspace(int c);



void print_box(BOX3D *box);
void print_box_oneline(BOX3D *box);
void print_point(char *result, POINT3D *pt,bool is3d);
void print_many_points(char *result, POINT3D *pt ,int npoints, bool is3d);
void swap(double *d1, double *d2);
int	numb_points_in_list(char	*str);
bool	parse_points_in_list(char	*str, POINT3D	*points, int32	max_points, bool *is3d);
bool	parse_points_in_list_exact(char	*str, POINT3D	*points, int32	max_points, bool *is3d);
int	find_outer_list_length(char *str);
bool	points_per_sublist( char *str, int32 *npoints, int32 max_lists);
char	*scan_to_same_level(char	*str);
int objects_inside_point(char *str);
int objects_inside_line(char *str);
int objects_inside_polygon(char *str);
int objects_inside_multipoint(char *str);
int objects_inside_multiline(char *str);
int objects_inside_multipolygon(char *str);
int objects_inside_collection(char *str);
int objects_inside(char *str);
bool parse_objects_inside_point(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d);
bool parse_objects_inside_multipoint(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d);
bool parse_objects_inside_line(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d);
bool parse_objects_inside_multiline(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d);
bool parse_objects_inside_polygon(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d);
bool parse_objects_inside_multipolygon(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d);
bool parse_objects(int32 *obj_size,char **objs,int32	*obj_types,int32 nobjs,char *str, int *offset, bool *is3d);
bool parse_objects_inside_collection(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d);
BOX3D	*bbox_of_point(POINT3D *pt);
BOX3D	*bbox_of_polygon(POLYGON3D *polygon);
BOX3D	*bbox_of_line(LINE3D *line);
BOX3D *union_box3d(BOX3D *a, BOX3D *b);
BOX3D	*bbox_of_geometry(GEOMETRY *geom);
GEOMETRY *make_bvol_geometry(BOX3D *box);
static bool box3d_ov(BOX3D *box1, BOX3D *box2);
bool is_same_point(POINT3D	*p1, POINT3D	*p2);
bool is_same_line(LINE3D	*l1, LINE3D	*l2);
bool is_same_polygon(POLYGON3D	*poly1, POLYGON3D	*poly2);
BOX	*convert_box3d_to_box(BOX3D *in);
double line_length2d(LINE3D *line);
double line_length3d(LINE3D *line);
double polygon_area2d_old(POLYGON3D *poly1);
double 	polygon_perimeter3d(POLYGON3D	*poly1);
double 	polygon_perimeter2d(POLYGON3D	*poly1);
int PIP( POINT3D *P, POINT3D *V, int n );
bool point_truely_inside(POINT3D	*point, BOX3D	*box);
int	compute_outcode( POINT3D *p, BOX3D *box);
bool lineseg_inside_box( POINT3D *P1, POINT3D *P2, BOX3D *box);
bool	linestring_inside_box(POINT3D *pts, int npoints, BOX3D *box);
bool polygon_truely_inside(POLYGON3D	*poly, BOX3D *box);
bool line_truely_inside( LINE3D *line, BOX3D 	*box);
void	translate_points(POINT3D *pt, int npoints,double x_off, double y_off, double z_off);


void  swap_char(char *a, char*b);
void	flip_endian_double(char	*dd);
void		flip_endian_int32(char 	*ii);

char	*to_wkb(GEOMETRY *geom, bool flip_endian);
char	*wkb_multipolygon(POLYGON3D	**polys,int numb_polys,int32 *size, bool flipbytes, char byte_order,bool use3d);
char	*wkb_polygon(POLYGON3D	*poly,int32 *size, bool flipbytes, char byte_order,bool use3d, char *mem);
char	*wkb_multiline(LINE3D **lines,int32 *size, int numb_lines, bool flipbytes, char byte_order,bool use3d);
char	*wkb_line(LINE3D *line,int32 *size, bool flipbytes, char byte_order,bool use3d, char *mem);
char	*wkb_point(POINT3D *pt,int32 *size, bool flipbytes, char byte_order, bool use3d);
char	*wkb_multipoint(POINT3D *pt,int32 numb_points,int32 *size, bool flipbytes, char byte_order,bool use3d);

char *to_wkb_collection(GEOMETRY *geom, bool flip_endian, int32 *size);
char	*to_wkb_sub(GEOMETRY *geom, bool flip_endian, int32 *wkb_size);

void decode_wkb_collection(char *wkb,int	*size);
void decode_wkb(char *wkb, int *size);
void dump_bytes( char *a, int numb);

//exposed to psql

Datum box3d_in(PG_FUNCTION_ARGS);
Datum box3d_out(PG_FUNCTION_ARGS);
Datum geometry_in(PG_FUNCTION_ARGS);
Datum geometry_out(PG_FUNCTION_ARGS);
Datum get_bbox_of_geometry(PG_FUNCTION_ARGS);
Datum get_geometry_of_bbox(PG_FUNCTION_ARGS);
Datum box3d_same(PG_FUNCTION_ARGS);
Datum geometry_overleft(PG_FUNCTION_ARGS);
Datum geometry_left(PG_FUNCTION_ARGS);
Datum geometry_right(PG_FUNCTION_ARGS);
Datum geometry_overright(PG_FUNCTION_ARGS);
Datum geometry_contained(PG_FUNCTION_ARGS);
Datum geometry_contain(PG_FUNCTION_ARGS);
Datum geometry_overlap(PG_FUNCTION_ARGS);
Datum geometry_same(PG_FUNCTION_ARGS);
Datum box3d_overlap(PG_FUNCTION_ARGS);
Datum box3d_overleft(PG_FUNCTION_ARGS);
Datum box3d_right(PG_FUNCTION_ARGS);
Datum box3d_contain(PG_FUNCTION_ARGS);
Datum geometry_union(PG_FUNCTION_ARGS);
Datum geometry_inter(PG_FUNCTION_ARGS);
Datum geometry_size(PG_FUNCTION_ARGS);
Datum length3d(PG_FUNCTION_ARGS);
Datum length2d(PG_FUNCTION_ARGS);
Datum area2d(PG_FUNCTION_ARGS);
Datum perimeter3d(PG_FUNCTION_ARGS);
Datum perimeter2d(PG_FUNCTION_ARGS);
Datum truly_inside(PG_FUNCTION_ARGS);

Datum geometry_lt(PG_FUNCTION_ARGS);
Datum geometry_gt(PG_FUNCTION_ARGS);
Datum geometry_eq(PG_FUNCTION_ARGS);

Datum npoints(PG_FUNCTION_ARGS);
Datum nrings(PG_FUNCTION_ARGS);
Datum mem_size(PG_FUNCTION_ARGS);
Datum numb_sub_objs(PG_FUNCTION_ARGS);
Datum summary(PG_FUNCTION_ARGS);
Datum translate(PG_FUNCTION_ARGS);

Datum wkb_XDR(PG_FUNCTION_ARGS);
Datum wkb_NDR(PG_FUNCTION_ARGS);

Datum force_2d(PG_FUNCTION_ARGS);
Datum force_3d(PG_FUNCTION_ARGS);

Datum combine_bbox(PG_FUNCTION_ARGS);



//for GIST index
typedef char* (*BINARY_UNION)(char*, char*, int*);
typedef float (*SIZE_BOX)(char*);
typedef Datum (*RDF)(PG_FUNCTION_ARGS);


PG_FUNCTION_INFO_V1(ggeometry_compress);
PG_FUNCTION_INFO_V1(ggeometry_union);
PG_FUNCTION_INFO_V1(ggeometry_picksplit);
PG_FUNCTION_INFO_V1(ggeometry_consistent);
PG_FUNCTION_INFO_V1(ggeometry_penalty);
PG_FUNCTION_INFO_V1(ggeometry_same);


GISTENTRY * ggeometry_compress(PG_FUNCTION_ARGS);
GEOMETRYKEY *ggeometry_union(PG_FUNCTION_ARGS);
GIST_SPLITVEC * ggeometry_picksplit(PG_FUNCTION_ARGS);
bool ggeometry_consistent(PG_FUNCTION_ARGS);
float * ggeometry_penalty(PG_FUNCTION_ARGS);
bool * ggeometry_same(PG_FUNCTION_ARGS);

static char * ggeometry_binary_union(char *r1, char *r2, int *sizep);
static float size_geometrykey( char *pk );
PG_FUNCTION_INFO_V1(ggeometry_inter);
Datum ggeometry_inter(PG_FUNCTION_ARGS);

/*
** Common rtree-function (for all ops)
*/
static char * rtree_union(bytea *entryvec, int *sizep, BINARY_UNION bu);
static float * rtree_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result, BINARY_UNION bu, SIZE_BOX sb);
static GIST_SPLITVEC * rtree_picksplit(bytea *entryvec, GIST_SPLITVEC *v, int keylen, BINARY_UNION bu, RDF interop, SIZE_BOX sb);
static bool rtree_internal_consistent(BOX *key, BOX *query, StrategyNumber strategy);

PG_FUNCTION_INFO_V1(rtree_decompress);
GISTENTRY * rtree_decompress(PG_FUNCTION_ARGS);



/*--------------------------------------------------------------------
 * Useful floating point utilities and constants.
 * from postgres's geo_decls.c 
 * EPSILON modified to be more "double" friendly
 *-------------------------------------------------------------------*/



// from contrib/cube/cube.c

#define max(a,b)		((a) >	(b) ? (a) : (b)) 
#define min(a,b)		((a) <= (b) ? (a) : (b))
#define abs(a)			((a) <	(0) ? (-a) : (a))


void swap_char(char	*a,char *b)
{
	char c;

	c = *a;
	*a=*b;
	*b=c;
}


void	flip_endian_double(char 	*d)
{
	swap_char(d+7, d);
	swap_char(d+6, d+1);
	swap_char(d+5, d+2);
	swap_char(d+4, d+3);
}

void		flip_endian_int32(char		*i)
{
	swap_char (i+3,i);
	swap_char (i+2,i+1);
}



//debug function - whats really in that BOX3D?
void print_box(BOX3D *box)
{
	printf("box is at %p\n",box);
	if (box == NULL)
	{
		printf ("	+ BOX IS NULL\n");
		return;	
	}
	printf("	+ LLB = [%g,%g,%g]\n", box->LLB.x, box->LLB.y,box->LLB.z);
	printf("	+ URT = [%g,%g,%g]\n", box->URT.x, box->URT.y,box->URT.z);
}

//debug function - whats really in that BOX3D?
void print_box_oneline(BOX3D *box)
{
	printf(" (%p) {",box);
	if (box == NULL)
	{
		printf ("BOX IS NULL}");
		return;	
	}
	printf("[%g,%g,%g] ", box->LLB.x, box->LLB.y,box->LLB.z);
	printf("[%g,%g,%g]}", box->URT.x, box->URT.y,box->URT.z);
}



//takes a point and sticks it at the end of a string
void print_point(char *result, POINT3D *pt,bool is3d)
{
	char	temp[MAX_DIGS_DOUBLE*3 +6];


	if ( (pt == NULL) || (result == NULL) )
		return;

	if (is3d)
		sprintf(temp, "%.15g %.15g %.15g", pt->x,pt->y,pt->z);
	else
		sprintf(temp, "%.15g %.15g", pt->x,pt->y);

	strcat(result,temp);

}

//result MUST BE null terminated (or result[0] = 0)
void print_many_points(char *result, POINT3D *pt ,int npoints, bool is3d)
{
	int	u;

	result += strlen(result);
	if (is3d)
	{
		for (u=0;u<npoints;u++)
		{
			if (u != 0)
			{
				result[0] = ',';
				result++;
			}
			result+=	sprintf(result,"%.15g %.15g %.15g", pt[u].x,pt[u].y,pt[u].z);
		}
	}
	else
	{
		for (u=0;u<npoints;u++)
		{
			if (u != 0)
			{
				result[0] = ',';
				result++;
			}
			result+=	sprintf(result,"%.15g %.15g", pt[u].x,pt[u].y);
		}
	}
}


//swap two doubles
void swap(double *d1, double *d2)
{
	double t;

	t = *d1;
	*d1 = *d2;
	*d2 = t;
}

void dump_bytes( char *a, int numb)
{
	int	t;

	for (t=0; t<numb; t++)
	{
		printf("	+ Byte #%i has value %i (%x)\n",t,a[t],a[t]);
	}
}

//returns how many points are in the first list in str
//
//  1. scan ahead looking for "("
//  2. find "," until hit a ")"
//  3. return number of points found
//	
// NOTE: doesnt actually parse the points, so if the 
//       str contains an invalid geometry, this could give
// 	   back the wrong answer.
//
// "(1 2 3, 4 5 6),(7 8, 9 10, 11 12 13)" => 2 (2nd list is not included)
int	numb_points_in_list(char	*str)
{
	bool	keep_going;
	int	points_found = 1; //no "," if only one point (and last point)
					// # points = 1 + numb of ","

	if ( (str == NULL) || (str[0] == 0) )
	{
		return 0;  //either null string or empty string
	}

	//look ahead for the "("

	str = strchr(str,'(') ;
	
	if ( (str == NULL) || (str[1] == 0) )  // str[0] = '(';
	{
		return 0;  //either didnt find "(" or its at the end of the string
	}

	keep_going = TRUE;
	while (keep_going) 
	{
		str=strpbrk(str,",)");  // look for a "," or ")"
		keep_going = (str != NULL);
		if (keep_going)  // found a , or )
		{
			if (str[0] == ')')
			{
				//finished
				return points_found;
			}
			else	//str[0] = ","
			{
				points_found++;
				str++; //move 1 char forward	
			}
		}
	}
	return points_found; // technically it should return an error.
}

//Actually parse the points in a list
//  "(1 2 3, 4 5 6),(7 8, 9 10, 11 12 13)" => 
//			points[0] = {1,2,3} and points[1] = {4,5,6}
//		(2nd list is not parsed)
//
// parse at most max_points (points should already have enough space)
// if any of the points are 3d, is3d is set to true, otherwise its untouched.
//	2d points have z=0.0
//
// returns true if everything parses okay, otherwise false

bool	parse_points_in_list(char	*str, POINT3D	*points, int32	max_points, bool *is3d)
{
	bool	keep_going;
	int	numb_found= 0;
	int	num_entities;

	if ( (str == NULL) || (str[0] == 0)  || (max_points <0) || (points == NULL) )
	{
		return FALSE;  //either null string or empty string or other problem
	}

	if (max_points == 0)
		return TRUE; //ask for nothing, get nothing

	//look ahead for the "("

	str = strchr(str,'(') ;
	
	if ( (str == NULL) || (str[1] == 0) )  // str[0] = '(';
	{
		return FALSE;  //either didnt find "(" or its at the end of the string
	}
	str++;  //move forward one char

	keep_going = TRUE;
	while (keep_going) 
	{
		//attempt to get the point
	      	num_entities = sscanf(str,"%le %le %le",  &(points[numb_found].x),
									      &(points[numb_found].y),
								     	      &(points[numb_found].z));

			if (num_entities !=3) 
			{
				if (num_entities !=2 )
				{
					elog(ERROR, "geom3d: parse_points_in_list() on invalid point");
					return FALSE; //error
				}
				else
				{
					points[numb_found].z = 0.0; //2d (only found x,y - set z =0.0)	
				}
			}	
			else
			{
				*is3d = TRUE; //found 3 entites (x,y,z)
			}
			numb_found++;



		str=strpbrk(str,",)");  // look for a "," or ")"
		if (str != NULL)
			str++;
		keep_going =  ( (str != NULL) && (str[0] != ')' ) && (numb_found < max_points)  );
	}
	return TRUE; 
}


//Actually parse the points in a list
//  "(1 2 3, 4 5 6),(7 8, 9 10, 11 12 13)" => 
//			points[0] = {1,2,3} and points[1] = {4,5,6}
//		(2nd list is not parsed)
//
// parse at most max_points (points should already have enough space)
// if any of the points are 3d, is3d is set to true, otherwise its untouched.
//	2d points have z=0.0
//
// returns true if everything parses okay, otherwise false
//
// THIS IS EXACTLY the same as parse_points_in_list(), but returns FALSE if 
// the number of points parsed is != max_points

bool	parse_points_in_list_exact(char	*str, POINT3D	*points, int32	max_points, bool *is3d)
{
	bool	keep_going;
	int	numb_found= 0;
	
	char	*end_of_double;

	if ( (str == NULL) || (str[0] == 0)  || (max_points <0) || (points == NULL) )
	{
		return FALSE;  //either null string or empty string or other problem
	}

	if (max_points == 0)
		return TRUE; //ask for nothing, get nothing

	//look ahead for the "("

	str = strchr(str,'(') ;
	
	if ( (str == NULL) || (str[1] == 0) )  // str[0] = '(';
	{
		return FALSE;  //either didnt find "(" or its at the end of the string
	}
	str++;  //move forward one char

	keep_going = TRUE;
	while (keep_going) 
	{
		//attempt to get the point

			//scanf is slow, so we use strtod()

			
			points[numb_found].x = strtod(str,&end_of_double);
			if (end_of_double == str)
			{
				return FALSE; //error occured (nothing parsed)
			}
			str = end_of_double;
			points[numb_found].y = strtod(str,&end_of_double);
			if (end_of_double == str)
			{
				return FALSE; //error occured (nothing parsed)
			}
			str = end_of_double;
			points[numb_found].z = strtod(str,&end_of_double); //will be zero if error occured
			if (!(end_of_double == str))
			{
				*is3d = TRUE; //found 3 entites (x,y,z)
			}
			str = end_of_double;
			numb_found++;



		str=strpbrk(str,",)");  // look for a "," or ")"
		if (str != NULL)
			str++;
		keep_going =  ( (str != NULL) && (str[0] != ')' ) && (numb_found < max_points)  );
	}
	return (numb_found == max_points); 
}




// Find the number of sub lists in a list
// ( (..),(..),(..) )  -> 3
// ( ( (..),(..) ), ( (..) )) -> 2
// ( ) -> 0
// scan through the list, for every "(", depth (nesting) increases by 1
//				  for every ")", depth (nesting) decreases by 1
// if find a "(" at depth 1, then there is a sub list
//
// example: 
//      "(((..),(..)),((..)))"
//depth  12333223332112333210
//        +           +         increase here

int	find_outer_list_length(char *str)
{
	int	current_depth = 0;
	int	numb_lists = 0;


	while ( (str != NULL) && (str[0] != 0) )
	{
		str=strpbrk(str,"()"); //look for "(" or ")"
		if (str != NULL)
		{
			if (str[0] == '(')
			{
				current_depth++;
				if (current_depth == 2)
					numb_lists ++;
			}
			if (str[0] == ')')
			{
				current_depth--;
				if (current_depth == 0)
					return numb_lists ;
			}
			str++;
		}
	}
	return numb_lists ; // probably should give an error
}


// Find out how many points are in each sublist, put the result in the array npoints[]
//  (for at most max_list sublists)
//
//  ( (L1),(L2),(L3) )  --> npoints[0] = points in L1,
//				    npoints[1] = points in L2,
//				    npoints[2] = points in L3
//
// We find these by, again, scanning through str looking for "(" and ")"
// to determine the current depth.  We dont actually parse the points.

bool	points_per_sublist( char *str, int32 *npoints, int32 max_lists)
{
	//scan through, noting depth and ","s

	int	current_depth = 0;
	int	current_list =-1 ;


	while ( (str != NULL) && (str[0] != 0) )
	{
		str=strpbrk(str,"(),");  //find "(" or ")" or ","
		if (str != NULL)
		{
			if (str[0] == '(')
			{
				current_depth++;
				if (current_depth == 2)
				{
					current_list ++;
					if (current_list >=max_lists)
						return TRUE;			// too many sub lists found
					npoints[current_list] = 1;
				}
				// might want to return an error if depth>2
			}
			if (str[0] == ')')
			{
				current_depth--;
				if (current_depth == 0)
					return TRUE ;
			}
			if (str[0] == ',')
			{
				if (current_depth==2)
				{
					npoints[current_list] ++;	
				}
			}

			str++;
		}
	}
	return TRUE ; // probably should give an error
}




//simple scan-forward to find the next "(" at the same level
//  ( (), (),(), ),(...
//                 + return this location
char	*scan_to_same_level(char	*str)
{

	//scan forward in string looking for at "(" at the same level
	// as the one its already pointing at

		int	current_depth = 0;
		bool  first_one=TRUE;


	while ( (str != NULL) && (str[0] != 0) )
	{
		str=strpbrk(str,"()");
		if (str != NULL)
		{
			if (str[0] == '(')
			{
				if (!(first_one))
				{
					if (current_depth == 0)
						return str;
				}
				else
					first_one = FALSE;  //ignore the first opening "("
				current_depth++;
			}
			if (str[0] == ')')
			{
				current_depth--;
			}

			str++;
		}
	}
	return str ; // probably should give an error
}


/*
 *  BOX3D_in - takes a string rep of BOX3D and returns internal rep
 *
 *  example:
 *     "BOX3D(x1 y1 z1,x2 y2 z2)"
 * or  "BOX3D(x1 y1,x2 y2)"   z1 and z2 = 0.0
 *
 *
 */
PG_FUNCTION_INFO_V1(box3d_in);
Datum box3d_in(PG_FUNCTION_ARGS)
{
	char	   *str = PG_GETARG_CSTRING(0);
	BOX3D	   *bbox = (BOX3D *) palloc(sizeof(BOX3D));
	bool		junk_bool;
	bool	   okay;
	int 		npoints;

	//verify that there are exactly two points

	//strip leading spaces
	while (isspace((unsigned char) *str))
		str++;
//printf( "box3d_in gets '%s'\n",str);

	if (strstr(str,"BOX3D") !=  str ) 
	{
		 elog(ERROR,"BOX3D parser - doesnt start with BOX3D");
		 pfree(bbox);
		 PG_RETURN_NULL();	
	}

	if ((npoints = numb_points_in_list(str)) != 2)
	{
//printf("npoints in this bbox is %i, should be 2\n",npoints);
		 elog(ERROR,"BOX3D parser - number of points should be exactly 2");
		 pfree(bbox);
		 PG_RETURN_NULL();
	}

	//want to parse two points, and dont care if they are 2d or 3d
	okay = 	parse_points_in_list(str, &(bbox->LLB), 2, &junk_bool);
	if (okay == 0)
	{
		elog(ERROR,"box3d: couldnt parse: '%s'\n",str);
		pfree(bbox);
		PG_RETURN_NULL() ;
	}	
	

	//validate corners so LLB is mins of x,y,z and URT is maxs x,y,z
	if (	bbox->LLB.x > bbox->URT.x)
	{
		swap ( &bbox->LLB.x , &bbox->URT.x ) ;
	}
	if (	bbox->LLB.y > bbox->URT.y)
	{
		swap ( &bbox->LLB.y , &bbox->URT.y ); 
	}
	if (	bbox->LLB.z > bbox->URT.z)
	{
		swap ( &bbox->LLB.z , &bbox->URT.z ) ;
	}

	PG_RETURN_POINTER(bbox);
}

/*
 *  Takes an internal rep of a BOX3D and returns a string rep.
 *
 *  example:
 *     "BOX3D(LLB.x LLB.y LLB.z, URT.x URT.y URT.z)"
 */
PG_FUNCTION_INFO_V1(box3d_out);
Datum box3d_out(PG_FUNCTION_ARGS)
{
	BOX3D  *bbox = (BOX3D *) PG_GETARG_POINTER(0);
	int size;
	char	*result;

	if (bbox == NULL)
	{
		result = palloc(5);
		strcat(result,"NULL");
		PG_RETURN_CSTRING(result);
	}

	size = MAX_DIGS_DOUBLE*6+5+2+4+5+1;
	result = (char *) palloc(size); //double digits+ "BOX3D"+ "()" + commas +null
	sprintf(result, "BOX3D(%.15g %.15g %.15g,%.15g %.15g %.15g)",
			bbox->LLB.x,bbox->LLB.y,bbox->LLB.z,
			bbox->URT.x,bbox->URT.y,bbox->URT.z);

	PG_RETURN_CSTRING(result);
}

/***************************************************************
 * these functions return the number of sub-objects inside a 
 * sub-portion of a string.
 * 
 * LINESTRING(), POLYGON(), POINT() allways have 1 sub object
 *
 * MULTIPOLYGON() MULTILINESTRING() MULTIPOINT() will have a variable number of sub-parts
 *           ex MULTIPOINT(1 1 1, 2 2 2,3 3 3,4 4 4) -->4 because there are 4 points in it
 *
 * GEOMETRYCOLLECTION() - need to look at each sub-object in the collection
 ***************************************************************/

//number of object inside a point
int objects_inside_point(char *str)
{
	return 1; //trivial points always have 1 point in them
}


//number of object inside a line
int objects_inside_line(char *str)
{
	return 1; //trivial lines always have 1 line in them
}

//number of object inside a polygon
int objects_inside_polygon(char *str)
{
	return 1; //trivial polygon always have 1 polygon in them
}


//bit more complicated - need to find out how many points are in this entity
int objects_inside_multipoint(char *str)
{
	return numb_points_in_list(str);
}

//bit more complicated - need to find out how many lines are in this entity
int objects_inside_multiline(char *str)
{
	return find_outer_list_length(str);
}


//bit more complicated - need to find out how many polygons are in this entity
int objects_inside_multipolygon(char *str)
{
	return find_outer_list_length(str);
}



//This one is more complicated, we have to look at each of the sub-object inside
// the collection.
//  result = sum of ( number of objects in each of the collection's sub-objects )
//
// EX:   'GEOMETRYCOLLECTION( POINT(1 2 3),
//	POINT (4 5 6) ),  
//	LINESTRING(3 4 5,1 2 3,5 6 0), 
//	MULTILINESTRING((3 4 5,1 2 3,5 6 0)),  
//	MULTIPOINT(1 2 3),  
//	MULTILINESTRING((3 4 5,1 2 3,5 6 0),(1 2 0,4 5 0,6 7 0,99 99 0)),
//	MULTIPOLYGON ((  (3 4 5,1 2 3,5 6),(1 2, 4 5, 6 7)  ), ( (11 11, 22 22, 33 33),(44 44, 55 55, 66 66) ) )'
//
//	# object = # of objs in POINT (1) + # of objs in POINT (1) +   # of objs in LINESTRING (1)
//			+ # of objs in MULTILINESTRING (1) + # of objs in MULTIPOINT  (1)
//			+ # of objs in MULTILINESTRING (1) + # of objs in  MULTIPOLYGON  (2)

int objects_inside_collection(char *str)
{
	int		tally = 0;
	int		sub_size = 0;

	//skip past the "geometrycollection" at begining of string
	str += 18; 
	if (strstr(str, "GEOMETRYCOLLECTION") != NULL)
		return -1; // collection inside collection; bad boy!

	//we scan to the first letter, then pass it to objects inside
 	// to find out how many sub-objects there are.  Then move to
 	// the next object


	while ( str != NULL)
	{
		//scan for a letter  (all objs start with M, P, or L
		str = strpbrk(str,"MPL"); //search for MULTI*, POLY*, POIN*
		if (str != NULL)
		{	
			//object found
			sub_size = objects_inside(str); //how many in this object
			if (sub_size == -1)
				return -1;//relay error

			tally += sub_size;	// running total

			//need to move to next item
			// by scanning past any letters
			str = strchr(str,'(');  //need to move to the next sub-object
		}
	}
	return tally;
}


// Find the # of sub-objects inside the geometry
// 	Just pass it off to the other functions
int objects_inside(char *str)
{

	char		*parenth;
	char		*loc;


	parenth = strchr(str,'(');
	if (parenth == NULL)
		return -1; //invalid string

	// look only at the begining of the string
	// the order of these "if" are important!

	if (  ((loc = strstr(str, "GEOMETRYCOLLECTION")) != NULL) && (loc < parenth) )
		return objects_inside_collection(str);

	// do multi before base types (MULTIPOINT has the string "POINT" in it)

	if (((loc = strstr(str,"MULTIPOINT")) != NULL) && (loc < parenth) )
		return objects_inside_multipoint(str);
	if (((loc = strstr(str,"MULTILINESTRING") )!= NULL) && (loc < parenth) )
		return objects_inside_multiline(str);
	if (((loc = strstr(str,"MULTIPOLYGON")) != NULL) && (loc < parenth) )
		return objects_inside_multipolygon(str);

	//base types 
	if (((loc = strstr(str,"POINT")) != NULL) && (loc < parenth) )
		return objects_inside_point(str);
	if (((loc = strstr(str,"LINESTRING")) != NULL) && (loc < parenth) )
		return objects_inside_line(str);
	if (((loc = strstr(str,"POLYGON")) != NULL) && (loc < parenth) )
		return objects_inside_polygon(str);

	return -1; //invalid

}


/****************************************************************************
 *  These functions actually parse the objects in the strings
 *  They all have arguments:
 * 	obj_size[]  -- size (in bytes) of each the sub-objects
 *    objs[][]    -- list of pointers to the actual objects
 *    obj_types   -- type of the sub-obj (see postgis.h)
 *    nobjs       -- max number of sub-obj
 *    str		-- actual string to parse from (start at begining)
 *    offset      -- which sub-object # are we dealing with
 *    is3d  	-- are any of the points in the object 3d?
 *
 *    basically, each of the functions will:
 *		+ parse str to find the information required
 *		+ allocate a new obj(s) and place it/them in objs to hold this info
 *		+ set the obj_size for this/these sub-object
 *		+ set the obj_type for this/these sub-object
 *		+ increase offset so the next objects will go in the correct place
 *		+ set is3d if there are any 3d points in the sub-object
 *		+ return FALSE if an error occured
 ****************************************************************************/


//parse a point - make a new POINT3D object
bool parse_objects_inside_point(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d)
{
	bool	result;

	if (*offset >= nobjs)
		return FALSE;     //too many objs found

	objs[*offset] = palloc (sizeof(POINT3D) );
	memset(objs[*offset], 0, sizeof(POINT3D) ); // zero memory
	obj_types[*offset] = POINTTYPE;
	obj_size[*offset] = sizeof(POINT3D);

	str = strchr(str,'(');
	if (str == NULL)
		return FALSE;  // parse error

//printf("about to parse the point\n");

	result = parse_points_in_list_exact(str, (POINT3D*) objs[*offset] , 1, is3d);
//printf("	+point parsed\n");

	*offset = *offset + 1;
	return (result); // pass along any parse error
}


//parse a multipoint object - offset may increase by >1
// multiple sub object may be created
bool parse_objects_inside_multipoint(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d)
{
	bool	result;
	int	npoints;
	int	t;
	POINT3D	*pts;

	// first find out how many points are in the multi-point
	npoints = objects_inside_multipoint(str);

	if (npoints <0)
		return FALSE; // error in parsing (should have been caught before)

	if (npoints ==0)
		return TRUE;  //no points, no nothing

	if (*offset >= nobjs)
		return FALSE;     //too many objs found dont do anything, relay error


	str = strchr(str,'(');
	if (str == NULL)
		return FALSE;

		//allocate all the new objects
	pts = palloc( sizeof(POINT3D) * npoints);
	for (t=0;t<npoints; t++)
	{
		objs[*offset + t] = palloc( sizeof(POINT3D) );
		memset(objs[*offset+t], 0, sizeof(POINT3D) ); // zero memory

		obj_types[*offset+t] = POINTTYPE;
		obj_size[*offset+t] = sizeof(POINT3D);
	}
	
	//actually do the parsing into a temporary list
	result = parse_points_in_list_exact(str, pts , npoints, is3d);


	if (!(result))
	{
		pfree(pts);
		return FALSE; //relay error
	}

	for(t=0;t<npoints;t++)
	{
		memcpy(objs[*offset+t], &pts[t], sizeof(POINT3D) );
	}

	pfree(pts);
	*offset = *offset + npoints;
	return (result);
}


//parse a linestring
bool parse_objects_inside_line(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d)
{
	bool	result;
	int	num_points;
	bool	add_point = FALSE; // only 1 point --> make it two on top of each other


	if (*offset >= nobjs)
		return FALSE;     //too many objs found

	//how many points are in this linestring?
	num_points = numb_points_in_list(str);

	if (num_points == 0)
		return FALSE;


	if (num_points == 1)	//not really a line
	{
		num_points  = 2;
		add_point = TRUE;
	}


	
							// -1 because object already has 1 point in it
		objs[*offset] = palloc (sizeof(LINE3D) + sizeof(POINT3D)*(num_points-1) );
		memset(objs[*offset], 0, sizeof(LINE3D) + sizeof(POINT3D)*(num_points-1)  ); // zero memory

	obj_types[*offset] = LINETYPE;
	obj_size[*offset] = sizeof(LINE3D) + sizeof(POINT3D)*(num_points-1) ;

	str = strchr(str,'(');
	if (str == NULL)
		return FALSE;

	( (LINE3D *) objs[*offset] )->npoints = num_points;

	if (add_point)
	{
		result = parse_points_in_list_exact(str, &(( (LINE3D*) objs[*offset]) ->points[0]) , num_points-1, is3d);
	}
	else
	{
		result = parse_points_in_list_exact(str, &(( (LINE3D*) objs[*offset]) ->points[0]) , num_points, is3d);
	}

	if (add_point)
	{	
		memcpy( &(( (LINE3D*) objs[*offset]) ->points[1])   , &(( (LINE3D*) objs[*offset]) ->points[0])  , sizeof(POINT3D) );
	}

	*offset = *offset + 1;
	return (result);
}



//parse a multi-line.  For each sub-linestring, call the parse_objects_inside_line() function
// 	'MULTILINESTRING((3 4 5,1 2 3,5 6 0),(1 2 0,4 5 0,6 7 0,99 99 0))'
// first scan to the second "(" - this is the begining of the first linestring.
// scan it, then move to the second linestring
bool parse_objects_inside_multiline(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d)
{
	bool	result;
	int	num_lines;
	int 	t;


		//how many lines are there?
	num_lines = objects_inside_multiline(str);


	if (num_lines <0)
		return FALSE;

	if (num_lines ==0)
		return TRUE;	//can be nothing
	

	if (*offset >= nobjs)
		return FALSE;     //too many objs found

	str= strchr(str,'('); //first one
	if (str != NULL)
		str++;
	str= strchr(str,'('); //2nd one

	if (str == NULL)
		return FALSE;  // couldnt find it

	for (t=0;t<num_lines; t++)
	{
			//pass along the work - it will update offset for us
		result = parse_objects_inside_line(obj_size,objs, obj_types,nobjs, str,offset,is3d);
		if (!(result))
			return FALSE;//relay error
		if (str != NULL)
			str++;  // go past the first '(' in the linestring
		else
			return FALSE; //parse error
		str = strchr(str,'('); // go to the beginning of the next linestring
	}

	return (TRUE);  //everything okay
}

//parse a polygon
//	POLYGON (  (3 4 5,1 2 3,5 6),(1 2, 4 5, 6 7)  )'


bool parse_objects_inside_polygon(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d)
{
	bool	result;
	int	num_rings;
	int32	*npoints;
	int32	sum_points,points_offset;
	int	t;
	int	size;
	POINT3D	*pts;
	POLYGON3D	*polygon;



	if (*offset >= nobjs)
		return FALSE;     //too many objs found

		//find out how many rings there are
	num_rings = find_outer_list_length(str);
	
	if (num_rings <1)
		return FALSE; // must have at least one ring

		//allocate space for how many points in each ring
	npoints = palloc( sizeof(int32) * num_rings);

		//find out how many points are in each sub-list (ring)
	result = points_per_sublist(str,npoints, num_rings);

		//how many points are inside this polygon in total?
	sum_points =0;
	for(t=0;  t< num_rings ; t++)
	{
		sum_points += npoints[t];
		if (npoints[t] <3)
		{
			elog(ERROR,"polygon has ring with <3 points.");
			pfree(npoints);
			return FALSE;
		}
	}
	
		//now we can allocate the structure


			// -1 because the POLYGON3D already has space for 1 ring with 1 point
			// 4 is for possible double-word alignment
	size = sizeof(POLYGON3D) + sizeof(int32)*(num_rings-1) +4+ sizeof(POINT3D)*(sum_points-1)  ;
	objs[*offset] = palloc (size);
		memset(objs[*offset], 0, size  ); // zero memory

	polygon = (POLYGON3D *) objs[*offset];
	obj_types[*offset] = POLYGONTYPE;
	obj_size[*offset] = size;

	str = strchr(str,'(');
	if (str == NULL)
	{
		pfree(npoints);
		return FALSE;
	}
	str++;

	polygon->nrings = num_rings;

		pts = (POINT3D *)  &(polygon->npoints[num_rings]); //pts[] is just past the end of npoints[]	

			// make sure its double-word aligned

		pts = (POINT3D *) MAXALIGN(pts);


//printf("POLYGON is at %p, size is %i, pts is at %p (%i)\n",objs[*offset],size, pts, ((void *) pts) - ((void *)objs[*offset]) );
//printf("npoints[0] is at %p, num_rings=%i\n", &(((POLYGON3D *) objs[*offset])->npoints[0] ), num_rings );

	points_offset = 0;
	for(t=0;t<num_rings;t++)	//for each ring
	{
		//set num points in this obj
		polygon->npoints[t] = npoints[t]; // set # pts in this ring
		
			//use exact because it will screw up otherwise
//printf("about to parse the actual points in ring #%i with %i points:\n",t,npoints[t]);
		result = parse_points_in_list_exact(str, 
					&pts[points_offset], 
					npoints[t], is3d);
//printf("	+done parsing points\n");

//printf("1st points in ring is: [%g,%g,%g]\n", pts[points_offset].x ,pts[points_offset].y ,pts[points_offset].z );

		if (!(result))
		{
			pfree(npoints);
			return FALSE;	//relay error
		}
		
		//first and last point are the same
		if (!  (FPeq(pts[points_offset].x , pts[points_offset + npoints[t]-1].x  )
			 &&FPeq(pts[points_offset].y , pts[points_offset + npoints[t]-1].y   )  
			 &&FPeq(pts[points_offset].z , pts[points_offset + npoints[t]-1].z   )    )   )
		{
			elog(ERROR,"polygon has ring where first point != last point");
			pfree(npoints);
			return FALSE; //invalid poly
		}


		points_offset += npoints[t]; //where to stick points in the next ring

		str = strchr(str,'('); // where is the next ring
		if (str == NULL)
		{
			pfree(npoints);
			return FALSE;  //relay parse error
		}
		str++;
	}

	pfree(npoints);
	*offset = *offset + 1;
	return (result);
}


//	MULTIPOLYGON ((  (3 4 5,1 2 3,5 6),(1 2, 4 5, 6 7)  ), ( (11 11, 22 22, 33 33),(44 44, 55 55, 66 66) ) )'
//	pass off work to parse_objects_inside_polygon()

bool parse_objects_inside_multipolygon(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d)
{
	bool	result;
	int	num_polys;
	int 	t;

	num_polys = objects_inside_multipolygon(str);  //how many polygons are inside me?

//printf("parse_objects_inside_multipolygon('%s')\n",str);

	if (num_polys <0)
		return FALSE;

	if (num_polys ==0)
		return TRUE;	//ask for nothing, get nothing	

	if (*offset >= nobjs)
		return FALSE;     //too many objs found

	str= strchr(str,'('); //first one (begining of list of polys)
	if (str != NULL)
	{
		str++;
		str= strchr(str,'('); //2nd one (beginning of 1st poly)
	}
	if (str == NULL)
		return FALSE;	//relay error

	for (t=0;t<num_polys; t++)
	{
		if (str == NULL)
			return FALSE; 	//relay error
//printf("parse_inside_multipolygon: parsing a polygon #%i\n",t);
		result = parse_objects_inside_polygon(obj_size,objs, obj_types,nobjs, str,offset,is3d);
		if (!(result))
			return FALSE;	//relay error

		//scan ahead to next "(" start of next polygon
		str = scan_to_same_level(str);
	}
	return (TRUE);
}


// relay work to the above functions
bool parse_objects(int32 *obj_size,char **objs,int32	*obj_types,int32 nobjs,char *str, int *offset, bool *is3d)
{
		char		*parenth;
		char		*loc;


//printf("parse_objects: parsing object offset=%i\n",*offset);

	if (str == NULL)
		return FALSE;

	parenth = strchr(str,'(');
	if  (parenth == NULL)
		return FALSE; //invalid string


	if (  ((loc = strstr(str, "GEOMETRYCOLLECTION")) != NULL) && (loc < parenth) )
		return parse_objects_inside_collection(obj_size,objs, obj_types, nobjs, str, offset, is3d);


	if (((loc = strstr(str,"MULTIPOINT")) != NULL) && (loc < parenth) )
		return parse_objects_inside_multipoint(obj_size,objs, obj_types, nobjs, str, offset, is3d);


	if (((loc = strstr(str,"MULTILINESTRING") )!= NULL) && (loc < parenth) )
		return parse_objects_inside_multiline(obj_size,objs, obj_types, nobjs, str, offset, is3d);


	if (((loc = strstr(str,"MULTIPOLYGON")) != NULL) && (loc < parenth) )
		return parse_objects_inside_multipolygon(obj_size,objs, obj_types, nobjs, str, offset, is3d);

	if (((loc = strstr(str,"POINT")) != NULL) && (loc < parenth) )
		return parse_objects_inside_point(obj_size,objs, obj_types, nobjs, str, offset, is3d);


	if (((loc = strstr(str,"LINESTRING")) != NULL) && (loc < parenth) )
		return parse_objects_inside_line(obj_size,objs, obj_types, nobjs, str, offset, is3d);

	if (((loc = strstr(str,"POLYGON")) != NULL) && (loc < parenth) )
		return parse_objects_inside_polygon(obj_size,objs, obj_types, nobjs, str, offset, is3d);


	return FALSE; //invalid
	
}


//look for each of the object inside a collection, and send the work to parse it along
bool parse_objects_inside_collection(int32 *obj_size,char **objs, int32 *obj_types, int32 nobjs, char *str, int *offset, bool* is3d)
{
	bool	result = FALSE;

	str += 18; //scan past "geometrycollection"
	if (strstr(str, "GEOMETRYCOLLECTION") != NULL)
		return FALSE; // collection inside collection; bad boy!

	while ( str != NULL)
	{
		//scan for a letter
		str = strpbrk(str,"MPL"); //search for MULTI*, POLY*, POIN*
		if (str != NULL)
		{	
			//object found
			result = parse_objects(obj_size,objs,obj_types,nobjs,str, offset, is3d);

			if (result == FALSE)
				return FALSE;//relay error

			//prepare to move to next item
			// by scanning past any letters
			str = strchr(str,'(');
		}
	}
	return result;
}

/**************************************************************
 * Find the bounding box3d of a sub-object
 **************************************************************/

//simple box encloses a point
BOX3D	*bbox_of_point(POINT3D *pt)
{
	BOX3D		*the_box = (BOX3D *) palloc (sizeof(BOX3D));

	the_box->LLB.x = pt->x;
	the_box->LLB.y = pt->y;
	the_box->LLB.z = pt->z;

	the_box->URT.x = pt->x;
	the_box->URT.y = pt->y;
	the_box->URT.z = pt->z;
	
	return the_box;
}

// box encloses all the points in all the rings
BOX3D	*bbox_of_polygon(POLYGON3D *polygon)
{
		BOX3D		*the_box ;
		int		numb_points =0,i;
		POINT3D	*pts,*pt;

	for (i=0; i<polygon->nrings; i++)
	{
		numb_points +=  polygon->npoints[i];
	}

	if (numb_points <1)
		return NULL;
	
	pts = (POINT3D *) ( (char *)&(polygon->npoints[polygon->nrings] )  );
	pts = (POINT3D *) MAXALIGN(pts);
	
	the_box = bbox_of_point(&pts[0]);

	for (i=1; i<numb_points;i++)
	{
		pt = &pts[i];

		if (pt->x < the_box->LLB.x)
			the_box->LLB.x = pt->x;

		if (pt->y < the_box->LLB.y)
			the_box->LLB.y = pt->y;

		if (pt->z < the_box->LLB.z)
			the_box->LLB.z = pt->z;

		if (pt->x > the_box->URT.x)
			the_box->URT.x = pt->x;

		if (pt->y > the_box->URT.y)
			the_box->URT.y = pt->y;

		if (pt->z > the_box->URT.z)
			the_box->URT.z = pt->z;

	}
	return the_box;
}

//box encloses points in line
BOX3D	*bbox_of_line(LINE3D *line)
{
	BOX3D		*the_box;
	POINT3D	*pt;
	int i;

	if (line->npoints <1)
	{
		return NULL;
	}
	the_box = bbox_of_point(&line->points[0]);

	for (i=1;i< line->npoints; i++)
	{
		pt = &line->points[i];
		if (pt->x < the_box->LLB.x)
			the_box->LLB.x = pt->x;

		if (pt->y < the_box->LLB.y)
			the_box->LLB.y = pt->y;

		if (pt->z < the_box->LLB.z)
			the_box->LLB.z = pt->z;

		if (pt->x > the_box->URT.x)
			the_box->URT.x = pt->x;

		if (pt->y > the_box->URT.y)
			the_box->URT.y = pt->y;

		if (pt->z > the_box->URT.z)
			the_box->URT.z = pt->z;
	}
	return the_box;
}

//merge box a and b (expand b)
//if b is null, new one is returned
// otherwise b is returned
BOX3D *union_box3d(BOX3D *a, BOX3D *b)
{
	if (a==NULL)
		return NULL;
	if (b==NULL)
	{
		b=(BOX3D*) palloc(sizeof(BOX3D));
		memcpy(b,a,sizeof(BOX3D) );
		return b;
	}


	if (a->LLB.x < b->LLB.x)
		b->LLB.x = a->LLB.x;
	if (a->LLB.y < b->LLB.y)
		b->LLB.y = a->LLB.y;
	if (a->LLB.z < b->LLB.z)
		b->LLB.z = a->LLB.z;

	if (a->URT.x > b->URT.x)
		b->URT.x = a->URT.x;
	if (a->URT.y > b->URT.y)
		b->URT.y = a->URT.y;
	if (a->URT.z > b->URT.z)
		b->URT.z = a->URT.z;
	return b;
}

BOX3D	*bbox_of_geometry(GEOMETRY *geom)
{
	int	i;
	int32	*offsets;
	char	*obj;
	BOX3D	*result=NULL;
	BOX3D	*a_box;

//printf("bbox_of_geometry(%p)\n", geom);

	if (geom->nobjs <1)
		return NULL;	//bbox of 0 objs is 0

	//where are the objects living?
	offsets = (int32 *) ( ((char *) &(geom->objType[0] ))+ sizeof(int32) * geom->nobjs ) ;


	//for each sub-object
	for(i=0; i<geom->nobjs;i++)
	{
		obj = (char *) geom +offsets[i] ;  


		if (geom->objType[i] == POINTTYPE)
		{
			a_box = bbox_of_point(  (POINT3D *) obj);
			result= union_box3d(a_box  ,result);

			if (a_box != NULL)
				pfree (a_box);
		}
		if (geom->objType[i] == LINETYPE)
		{
			a_box = bbox_of_line(  (LINE3D *) obj);
			result = union_box3d(a_box  ,result);
			if (a_box != NULL)
				pfree (a_box);
		}
		if (geom->objType[i] == POLYGONTYPE)
		{
			a_box = bbox_of_polygon(  (POLYGON3D *) obj);
			result =union_box3d(a_box  ,result);
			if (a_box != NULL)
				pfree (a_box);
		}	
	}
	return result;
}



// Parse a geometry object
//
// 1. Find information on how many sub-object there are
// 2. create a group of sub-object and populate them
// 3. make a large structure & populate with all the above info.
// 
//
PG_FUNCTION_INFO_V1(geometry_in);
Datum geometry_in(PG_FUNCTION_ARGS)
{
		char	   *str = PG_GETARG_CSTRING(0);
		GEOMETRY	*geometry;
		BOX3D		*the_bbox;
		int32 		nobjs;
		char		**objs;
		int32		*obj_types;
		int32		*obj_size;
		int		entire_size;
		bool		is3d = FALSE;
		bool		okay;
		int		next_loc;
		int32		*offsets;
		int		t;
		int		obj_parse_offset;



		//trim white
	while (isspace((unsigned char) *str))
		str++;

//printf("geometry_in got string ''\n");

	nobjs=	objects_inside(str);  //how many sub-objects
	if (nobjs <=0)	//dont allow zero-object items
	{
		elog(ERROR,"couldnt parse objects in GEOMETRY\n");
		PG_RETURN_NULL() ;
	}

//printf("geometry_in determins that there are %i sub-objects\n",nobjs);

	//allocate enough space for info structures
	
	
	objs = palloc(sizeof( char *) * nobjs);
	memset(objs, 0, sizeof( char *) * nobjs);

	obj_types = palloc(sizeof(int32) * nobjs);
	memset(obj_types, 0, sizeof(int32) * nobjs);

	obj_size  = palloc(sizeof(int32) * nobjs);
	memset(obj_size, 0, sizeof(int32) * nobjs);

	obj_parse_offset = 0; //start populating obj list at beginning


//printf("	+about to parse the objects\n");


	okay = parse_objects(obj_size,objs,obj_types,nobjs,str,&obj_parse_offset , &is3d);

//printf("	+finished parsing object\n");

	if (!(okay) || (obj_parse_offset != nobjs ) )
	{
		//free
		for (t =0; t<nobjs; t++)
		{
			if (objs[t] != NULL)
				pfree (objs[t]);
		}
		pfree(objs); pfree(obj_types); pfree(obj_size);
		elog(ERROR,"couldnt parse object in GEOMETRY\n");
		PG_RETURN_NULL();
	}

	entire_size = sizeof(GEOMETRY);

 	for (t =0; t<nobjs; t++)
	{
		//do they all have an obj_type?
		if ( (obj_types[t] == 0) || (objs[t] == NULL) )
		{
			okay = FALSE;
		}
		entire_size += obj_size[t];  //total space of the objects

		//size seem reasonable?
		if (obj_size[t] <1)
		{
			okay = FALSE;
		}

	}
	if (!(okay) )
	{
		for (t =0; t<nobjs; t++)
		{
			if (objs[t] != NULL)
				pfree (objs[t]);
		}
		pfree(objs); pfree(obj_types); pfree(obj_size);
		elog(ERROR,"couldnt parse object in GEOMETRY\n");
		PG_RETURN_NULL();
	}

		//size = object size + space for object type & offset array and possible double-word boundary stuff
	entire_size = entire_size + sizeof(int32) *nobjs*2 + nobjs*4; 
	geometry = (GEOMETRY *) palloc (entire_size);

	geometry->size = entire_size;

	//set collection type.  Need to differentiate between
 	//  geometrycollection and (say) multipoint 

	if (   strstr(str, "GEOMETRYCOLLECTION") != NULL  )
		geometry->type = COLLECTIONTYPE;
	else
	{
		if (   strstr(str, "MULTI") != NULL  )
		{
			if (obj_types[0] == POINTTYPE)
				geometry->type = MULTIPOINTTYPE;
			if (obj_types[0] == LINETYPE)
				geometry->type = MULTILINETYPE;
			if (obj_types[0] == POLYGONTYPE)
				geometry->type = MULTIPOLYGONTYPE;
		}
		else
		{
			geometry->type = obj_types[0]; //single object
		}
	}

	geometry->is3d = is3d;    //any 3d points found anywhere?
	geometry->nobjs = nobjs; // sub-objects


	//copy in type and length info

			//where to put objects.  next_loc will point to (bytes forward) &objData[0]
		next_loc = ( (char *) &(geometry->objType[0] ) - (char *) geometry);
		next_loc += sizeof(int32) * 2* nobjs;

		next_loc = MAXALIGN(next_loc);
		


		//where is the offsets array
		offsets = (int32 *) ( ((char *) &(geometry->objType[0] ))+ sizeof(int32) * nobjs ) ;


//printf("structure is at %p, offsets is at %p, next_loc = %i\n",geometry, offsets, next_loc);

	for (t=0; t<nobjs; t++)	//for each object
	{
		geometry->objType[t] = obj_types[t];  //set its type

		offsets[t] = next_loc;	//where is it
//printf("copy %i bytes from %p to %p\n",	obj_size[t] ,  objs[t],(char *) geometry + next_loc);
		memcpy( (char *) geometry + next_loc, objs[t], obj_size[t] );	//put sub-object into object
		pfree(objs[t]); // free the original object (its redundant)

		next_loc += obj_size[t];	//where does the next object go?

		next_loc =  MAXALIGN(next_loc);


	}

	//free temporary structures
	pfree(objs); pfree(obj_types); pfree(obj_size);


	//calculate its bounding volume
	the_bbox = bbox_of_geometry(geometry);

	if (the_bbox != NULL)
	{
		memcpy( &geometry->bvol, the_bbox, sizeof(BOX3D) );
		pfree(the_bbox);
	}
	

//printf("returning from geometry_in, nobjs = %i\n", nobjs);
	PG_RETURN_POINTER(geometry);
}



//Take internal rep of geometry, output string
PG_FUNCTION_INFO_V1(geometry_out);
Datum geometry_out(PG_FUNCTION_ARGS)
{
	char	*result;
	GEOMETRY  *geometry = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int		t,u;
	int32		*offsets;
	char		*obj;
	POINT3D	*pts;
	POLYGON3D	*polygon;
	LINE3D	*line;
	POINT3D	*point;
	
	int		pt_off,size;
	bool		briefmode= TRUE; // ie. dont print out "POINT3D(..)", just print out the ".." part
	bool 		is3d;
	bool		first_sub_obj = TRUE;
	bool		multi_obj = FALSE;
	int		mem_size,npts;


printf("in geom_out(%p)\n",geometry);

	size = 30;	//just enough to put in object type
	result = (char *) palloc(30); mem_size= 30;       //try to limit number of repalloc()s

	if (geometry->type == BBOXONLYTYPE)
	{
		sprintf(result,"BOX3D ONLY - NOT useable");
	}


	if (geometry->type == POINTTYPE)
		sprintf(result, "POINT(" );
	if (geometry->type == LINETYPE)
		sprintf(result, "LINESTRING" );
	if (geometry->type == POLYGONTYPE)
		sprintf(result, "POLYGON" );

	if (geometry->type == MULTIPOINTTYPE)
		sprintf(result, "MULTIPOINT(" );
	if (geometry->type == MULTILINETYPE)
		sprintf(result, "MULTILINESTRING(" );
	if (geometry->type == MULTIPOLYGONTYPE)
		sprintf(result, "MULTIPOLYGON(" );

	if  ( (geometry->type == MULTIPOINTTYPE) || (geometry->type == MULTILINETYPE) || (geometry->type == MULTIPOLYGONTYPE))
	{
		multi_obj = TRUE;	//multiple sub-object need to be put into one object
	}

	if (geometry->type == COLLECTIONTYPE)
	{
		sprintf(result, "GEOMETRYCOLLECTION(" );
		briefmode = FALSE;
		multi_obj = FALSE;
	}
	
		//where are the objects?
	offsets = (int32 *) ( ((char *) &(geometry->objType[0] ))+ sizeof(int32) * geometry->nobjs ) ;

	is3d = geometry->is3d;

	
	for(t=0;t<geometry->nobjs; t++)  //for each object
	{
		obj = (char *) geometry +offsets[t] ;  

		if (geometry->objType[t] == 1)   //POINT
		{
			point = (POINT3D *) obj;
			if (briefmode)	//dont need to put in "POINT("
			{
				size +=MAX_DIGS_DOUBLE*3 + 2 +3  ;
				result = repalloc(result, size );  //make memory bigger
				if (!(first_sub_obj))
				{
					strcat(result,",");	
				}
				else
				{
					first_sub_obj = FALSE;
				}
				print_point(result, point,is3d);  //render point
				if (t == (geometry->nobjs-1)) 
					strcat(result,")");	// need to close object?
			}
			else
			{
				size +=MAX_DIGS_DOUBLE*3 + 2 +3 +7  ;
				result = repalloc(result, size  );
				strcat(result, "POINT(");
				print_point(result, point,is3d);
				strcat(result, ")");
				if (t != (geometry->nobjs -1) )
					strcat(result,",");
			}

		}
		if (geometry->objType[t] == 2)	//LINESTRING
		{
			line = (LINE3D *) obj;
			if (briefmode)
			{
				size +=(MAX_DIGS_DOUBLE*3+5)*line->npoints +3;
				result = repalloc(result, size );
				if (!(first_sub_obj))
				{
					strcat(result,",");	
				}
				else
				{
					first_sub_obj = FALSE;
				}

				strcat(result,"(");
				print_many_points(result, &line->points[0],line->npoints , is3d);
				strcat(result,")");
				if ((t == (geometry->nobjs-1)) && multi_obj )  //close object?
					strcat(result,")");
			}
			else
			{
				size +=(MAX_DIGS_DOUBLE*3+5)*line->npoints +12+3;
				result = repalloc(result, size );
				strcat(result, "LINESTRING(");

				//optimized for speed

				print_many_points(result, &line->points[0],line->npoints , is3d);
				strcat(result,")");
				if (t != (geometry->nobjs -1) )
					strcat(result,",");

			}
		}
		if (geometry->objType[t] == 3)		//POLYGON
		{
				polygon = (POLYGON3D *) obj;
				if (!(briefmode))
				{
					size += 7 + polygon->nrings *3+9;
					result = repalloc(result, size );
					strcat(result,"POLYGON");
				}	
				else
				{
					size += 7 + polygon->nrings *3;
					result = repalloc(result, size );
				}
				
				//need to separate objects?
				if (!(first_sub_obj))
				{
					strcat(result,",");	
				}
				else
				{
					first_sub_obj = FALSE;
				}

				
				strcat(result,"(");	//begin poly

				pt_off = 0;	//where is the first point in this ring?

					//where are the points
				pts = (POINT3D *) ( (char *)&(polygon->npoints[polygon->nrings] )  );
				pts = (POINT3D *) MAXALIGN(pts);

				
				npts = 0;
				for (u=0; u< polygon->nrings ; u++)  
					npts += polygon->npoints[u];

				size += (MAX_DIGS_DOUBLE*3+3) * npts + 5* polygon->nrings;
				result = repalloc(result, size );
				
				for (u=0; u< polygon->nrings ; u++)  //for each ring
				{
//printf("making ring #%i in obj, %i\n",u,t);
					if (u!=0)
						strcat(result,","); //separate rings

					strcat(result,"(");	//begin ring
					print_many_points(result, &pts[pt_off] ,polygon->npoints[u], is3d);

					pt_off = pt_off + polygon->npoints[u]; //where is first point of next ring?
					strcat(result,")");	//end ring
				}
				strcat(result,")"); //end poly
				if ((t == (geometry->nobjs-1)) && multi_obj )
					strcat(result,")");
				if ((!(briefmode)) && (t != (geometry->nobjs -1) ))
					strcat(result,",");
		}
		if (!(briefmode))
		{
			first_sub_obj = TRUE;
		}
		
	}

	if (!(briefmode))
		strcat(result,")");
	PG_RETURN_CSTRING(result);
}


/*****************************************************
 * conversion routines
 *****************************************************/





//get bvol inside a geometry
PG_FUNCTION_INFO_V1(get_bbox_of_geometry);
Datum get_bbox_of_geometry(PG_FUNCTION_ARGS)
{
	GEOMETRY  *geom = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));

	BOX3D	 *result;
	
	//make a copy of it

	result = palloc ( sizeof(BOX3D) );
	memcpy(result,&geom->bvol, sizeof(BOX3D) );


	PG_RETURN_POINTER(result);
}


//make a fake geomery with just it a bvol
PG_FUNCTION_INFO_V1(get_geometry_of_bbox);
Datum get_geometry_of_bbox(PG_FUNCTION_ARGS)
{
	BOX3D  *bbox = (BOX3D *) PG_GETARG_POINTER(0);
	GEOMETRY  *geom   ;
	
	//make a copy of it

	geom = palloc ( sizeof(GEOMETRY) );
	geom->size = sizeof(GEOMETRY);
	geom->type = BBOXONLYTYPE;
	geom->nobjs = -1;
	memcpy(&(geom->bvol),bbox, sizeof(BOX3D) );


	PG_RETURN_POINTER(geom);
}

//given a bvol, make a "fake" geometry that contains it (for indexing)
GEOMETRY *make_bvol_geometry(BOX3D *box)
{
	return (GEOMETRY *) DatumGetPointer(DirectFunctionCall1(get_geometry_of_bbox,PointerGetDatum(box) ) );
}

//   ***********************************
//          GEOMETRY indexing (operator) fns
//	NOTE:
//		These work on the bbox of the geometry, not the 
//		individual components (except same)
//   ***********************************


PG_FUNCTION_INFO_V1(box3d_same);
Datum box3d_same(PG_FUNCTION_ARGS)
{
	BOX3D		   *box1 = (BOX3D *)  PG_GETARG_POINTER(0);
	BOX3D		   *box2 = (BOX3D *) PG_GETARG_POINTER(1);

	bool		result;

	result = (	   FPeq(box1->LLB.x	, box2->LLB.x) &&
				   FPeq(box1->LLB.y	, box2->LLB.y) &&
				   FPeq(box1->URT.x	, box2->URT.x) &&
				   FPeq(box1->URT.y	, box2->URT.y)
				);
	PG_RETURN_BOOL(result);
}




PG_FUNCTION_INFO_V1(geometry_overleft);
Datum geometry_overleft(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	bool	result;

//printf("in geometry_overleft\n");
//print_box(&geom1->bvol);
//print_box(&geom2->bvol);
	result = FPle(geom1->bvol.URT.x, geom2->bvol.URT.x);

	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(geometry_left);
Datum geometry_left(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	bool	result;

//printf("in geometry_left\n");

	result = FPlt(geom1->bvol.URT.x, geom2->bvol.LLB.x);


	PG_RETURN_BOOL(result);
}


PG_FUNCTION_INFO_V1(geometry_right);
Datum geometry_right(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	bool	result;
//printf("in geometry_right\n");

	result = FPgt(geom1->bvol.LLB.x, geom2->bvol.URT.x);
	
	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(geometry_overright);
Datum geometry_overright(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	bool	result;
//printf("in geometry_overright\n");


	result = FPge(geom1->bvol.LLB.x, geom2->bvol.LLB.x);
	


	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(geometry_contained);
Datum geometry_contained(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	bool	result;

//printf("in geometry_contained\n");


	result = FPle(geom1->bvol.URT.x, geom2->bvol.URT.x) &&
				   FPge(geom1->bvol.LLB.x, geom2->bvol.LLB.x) &&
				   FPle(geom1->bvol.URT.y, geom2->bvol.URT.y) &&
				   FPge(geom1->bvol.LLB.y, geom2->bvol.LLB.y);
	
//printf("returning geometry_contained\n");
	PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(geometry_contain);
Datum geometry_contain(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	bool	result;
//printf("in geometry_contain\n");

	result = FPge(geom1->bvol.URT.x, geom2->bvol.URT.x) &&
				   FPle(geom1->bvol.LLB.x, geom2->bvol.LLB.x) &&
				   FPge(geom1->bvol.URT.y, geom2->bvol.URT.y) &&
				   FPle(geom1->bvol.LLB.y, geom2->bvol.LLB.y);
	


	PG_RETURN_BOOL(result);
}

static bool box3d_ov(BOX3D *box1, BOX3D *box2)
{
	bool	result;

	result = 

	/*overlap in x*/
		 ((    FPge(box1->URT.x, box2->URT.x) &&
			 FPle(box1->LLB.x, box2->URT.x)) ||
			(FPge(box2->URT.x, box1->URT.x) &&
			 FPle(box2->LLB.x, box1->URT.x)))
	&&
	/*overlap in y*/
	((FPge(box1->URT.y, box2->URT.y) &&
	  FPle(box1->LLB.y, box2->URT.y)) ||
	 (FPge(box2->URT.y, box1->URT.y) &&
	  FPle(box2->LLB.y, box1->URT.y)));

//printf("box3d_ov about to return %i\n",(int) result);	
	return result;
}


PG_FUNCTION_INFO_V1(geometry_overlap);
Datum geometry_overlap(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
//printf("in geometry_overlap\n");

	PG_RETURN_BOOL(box3d_ov(&(geom1->bvol), &(geom2->bvol)));
}


bool is_same_point(POINT3D	*p1, POINT3D	*p2)
{
	return ( FPeq(p1->x, p2->x ) && FPeq(p1->y, p2->y ) &&FPeq(p1->z, p2->z ) );
}

bool is_same_line(LINE3D	*l1, LINE3D	*l2)
{
	int i;
		//lines are directed, so all the points should be the same

	if (l1->npoints != l2->npoints)
		return FALSE; //simple case, not the same # of points

	for (i=0;i<l2->npoints; i++)
	{
		if (!( is_same_point( &l1->points[i], &l2->points[i] ) ) )
		{
			return FALSE;
		}
	}
	return TRUE;
}



PG_FUNCTION_INFO_V1(geometry_lt);
Datum geometry_lt(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	if  (geom1->bvol.LLB.x < geom2->bvol.LLB.x)  
		PG_RETURN_BOOL(TRUE);

	if ( FPeq(geom1->bvol.LLB.x , geom2->bvol.LLB.x)   )
	{ 
		if ( geom1->bvol.LLB.y < geom2->bvol.LLB.y )  
			PG_RETURN_BOOL(TRUE);

		if ( FPeq(geom1->bvol.LLB.y , geom2->bvol.LLB.y)   )
		{ 
			if ( geom1->bvol.LLB.z < geom2->bvol.LLB.z )  
				PG_RETURN_BOOL(TRUE);
			else
				PG_RETURN_BOOL(TRUE);
		}
		else
		{
			PG_RETURN_BOOL(FALSE);
		}

	}
	else
	{
		PG_RETURN_BOOL(FALSE);
	}
}

PG_FUNCTION_INFO_V1(geometry_gt);
Datum geometry_gt(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	if  (geom1->bvol.LLB.x > geom2->bvol.LLB.x)  
		PG_RETURN_BOOL(TRUE);

	if ( FPeq(geom1->bvol.LLB.x , geom2->bvol.LLB.x)   )
	{ 
		if ( geom1->bvol.LLB.y > geom2->bvol.LLB.y )  
			PG_RETURN_BOOL(TRUE);

		if ( FPeq(geom1->bvol.LLB.y , geom2->bvol.LLB.y)   )
		{ 
			if ( geom1->bvol.LLB.z > geom2->bvol.LLB.z )  
				PG_RETURN_BOOL(TRUE);
			else
				PG_RETURN_BOOL(TRUE);
		}
		else
		{
			PG_RETURN_BOOL(FALSE);
		}

	}
	else
	{
		PG_RETURN_BOOL(FALSE);
	}
}

PG_FUNCTION_INFO_V1(geometry_eq);
Datum geometry_eq(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	if  ( FPeq(geom1->bvol.LLB.x , geom2->bvol.LLB.x)  && FPeq(geom1->bvol.LLB.y , geom2->bvol.LLB.y)  && FPeq(geom1->bvol.LLB.z , geom2->bvol.LLB.z)  )
		PG_RETURN_BOOL(TRUE);
	else
		PG_RETURN_BOOL(FALSE);
}







//order of points in a ring IS important
//order of rings is also important
bool is_same_polygon(POLYGON3D	*poly1, POLYGON3D	*poly2)
{
	int	i;
	int	numb_points;
	POINT3D	*pts1, *pts2;


	if (poly1->nrings != poly2->nrings)
		return FALSE;
	
	numb_points = 0;
	for (i=0; i< poly1->nrings; i++)
	{
		if (poly1->npoints[i] != poly2->npoints[i])
			return FALSE;
		numb_points += poly1->npoints[i];
	}

	pts1 = (POINT3D *) ( (char *)&(poly1->npoints[poly1->nrings] )  );
	pts1 = (POINT3D *) MAXALIGN(pts1);
	

	pts2 = (POINT3D *) ( (char *)&(poly2->npoints[poly2->nrings] )  );
	pts2 = (POINT3D *) MAXALIGN(pts2);

	for (i=0; i< numb_points; i++)
	{
		if (!( is_same_point( &pts1[i], &pts2[i] ) ) )
		{
			return FALSE;
		}

	}
	return TRUE;
}



//FIXED: 
// geom1 same as geom2
//  iff  
//	+ have same type
//	+ have same # objects
//	+ have same bvol
//	+ each object in geom1 has a corresponding object in geom2 (see above)
PG_FUNCTION_INFO_V1(geometry_same);
Datum geometry_same(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));

	bool				result;
	int				i,j;
	bool				*already_hit;
	int32				type1,type2;
	int32				*offsets1,*offsets2;
	char				*o1,*o2;


//printf("in geometry_same\n");

	//This was removed because this doesnt actually refer to the geometry, but how it is represented
	// to the user
	//easy tests
	//if (geom1->type != geom2-> type)
	//	PG_RETURN_BOOL(FALSE);

	

	if (geom1->nobjs != geom2->nobjs)
		PG_RETURN_BOOL(FALSE);	

	result = DatumGetBool(DirectFunctionCall2(box3d_same,
								PointerGetDatum(&geom1->bvol), 
								PointerGetDatum(&geom2->bvol)   )     );

	if (result == FALSE)
		PG_RETURN_BOOL(FALSE);

	if (geom1->nobjs<1)
		return FALSE; // no objects (probably a BOXONLYTYPE) and bvols dont match

//	printf(" 	+have to do it the hard way");

	already_hit = (bool *) palloc (geom2->nobjs * sizeof(bool) );
	memset(already_hit, 0, geom2->nobjs * sizeof(bool) ); //all false
	

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;
	offsets2 = (int32 *) ( ((char *) &(geom2->objType[0] ))+ sizeof(int32) * geom2->nobjs ) ;


	//now have to do a scan of each object

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];

		for(i=0;i<geom1->nobjs; i++)		//for each object in geom2
		{
			o2 = (char *) geom2 +offsets2[i] ;  
			type2=  geom2->objType[j];
			
			if (  (type1 == type2) && (!(already_hit[i])) )
			{
				//see if they match
				if (type1 == POINTTYPE)
				{
					if (is_same_point((POINT3D *) o1,(POINT3D *) o2))
					{
						already_hit[i] = TRUE;
						break;
					} 
				}
				if (type1 == LINETYPE)
				{
					if (is_same_line((LINE3D *) o1,(LINE3D *) o2))
					{
						already_hit[i] = TRUE;
						break;
					} 

				}
				if (type2 == POLYGONTYPE)
				{
					if (is_same_polygon((POLYGON3D *) o1,(POLYGON3D *) o2))
					{
						already_hit[i] = TRUE;
						break;
					} 
				}	
			}
		}
		if (i == geom1->nobjs)
			PG_RETURN_BOOL(FALSE);	 // couldnt find match
	}
	PG_RETURN_BOOL(TRUE);
}

/*******************************************************
 *Box3d routines
 *******************************************************/



/*		box_overlap		-		does box1 overlap box2?
 */
PG_FUNCTION_INFO_V1(box3d_overlap);
Datum box3d_overlap(PG_FUNCTION_ARGS)
{
	BOX3D		   *box1 = (BOX3D *) PG_GETARG_POINTER(0);
	BOX3D		   *box2 = (BOX3D *) PG_GETARG_POINTER(1);

	PG_RETURN_BOOL(box3d_ov(box1, box2));
}

/*		box_overleft	-		is the right edge of box1 to the left of
 *								the right edge of box2?
 *
 *		This is "less than or equal" for the end of a time range,
 *		when time ranges are stored as rectangles.
 */
PG_FUNCTION_INFO_V1(box3d_overleft);
Datum box3d_overleft(PG_FUNCTION_ARGS)
{
	BOX3D		   *box1 = (BOX3D *) PG_GETARG_POINTER(0);
	BOX3D		   *box2 = (BOX3D *) PG_GETARG_POINTER(1);

	bool	result;

	result = FPle(box1->URT.x, box2->LLB.x);

//printf("box3d_overleft about to return %i\n",(int) result);	

	PG_RETURN_BOOL(result);
}

/*		box_right		-		is box1 strictly right of box2?
 */
PG_FUNCTION_INFO_V1(box3d_right);
Datum box3d_right(PG_FUNCTION_ARGS)
{
	BOX3D		   *box1 = (BOX3D *) PG_GETARG_POINTER(0);
	BOX3D		   *box2 = (BOX3D *) PG_GETARG_POINTER(1);

	bool	result;

	result = FPgt(box1->LLB.x, box2->URT.x);
//printf("box3d_rightabout to return %i\n",(int) result);	

	PG_RETURN_BOOL(result);
}

/*		box_contain		-		does box1 contain box2?
 */
PG_FUNCTION_INFO_V1(box3d_contain);
Datum box3d_contain(PG_FUNCTION_ARGS)
{
	BOX3D		   *box1 = (BOX3D *) PG_GETARG_POINTER(0);
	BOX3D		   *box2 = (BOX3D *) PG_GETARG_POINTER(1);


bool	result;

	result = FPge(box1->URT.x, box2->URT.x) &&
				   FPle(box1->LLB.x, box2->LLB.x) &&
				   FPge(box1->URT.y, box2->URT.y) &&
				   FPle(box1->LLB.y, box2->LLB.y);
	//printf("box3d_contain about to return %i\n",(int) result);	


	PG_RETURN_BOOL(result);
}


/******************************************************
 * RTREE index requires these 3 functions
 ******************************************************/

PG_FUNCTION_INFO_V1(geometry_union);
Datum geometry_union(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 =  (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 =  (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
	GEOMETRY		   *result = (GEOMETRY *) palloc(sizeof(GEOMETRY) );
	BOX3D				*n;


	result->size = sizeof(GEOMETRY);
	result->type = BBOXONLYTYPE;
	result->nobjs = -1;


	n = &result->bvol;

	//cheat - just change the bbox

        n->URT.x = max(geom1->bvol.URT.x, geom2->bvol.URT.x);
        n->URT.y = max(geom1->bvol.URT.y, geom2->bvol.URT.y);
        n->LLB.x = min(geom1->bvol.LLB.x, geom2->bvol.LLB.x);
        n->LLB.y = min(geom1->bvol.LLB.y, geom2->bvol.LLB.y);

	
        n->URT.z = max(geom1->bvol.URT.z, geom2->bvol.URT.z);
        n->LLB.z = min(geom1->bvol.LLB.z, geom2->bvol.LLB.z);

	  PG_FREE_IF_COPY(geom1, 0);
        PG_FREE_IF_COPY(geom2, 1);

	PG_RETURN_POINTER(result);
}

PG_FUNCTION_INFO_V1(geometry_inter);
Datum geometry_inter(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 =  (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		   *geom2 =  (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
	GEOMETRY		   *result = (GEOMETRY *) palloc(sizeof(GEOMETRY) );


	result->size = sizeof(GEOMETRY);
	result->type = BBOXONLYTYPE;
	result->nobjs = -1;

		result->bvol.URT.x = min(geom1->bvol.URT.x, geom2->bvol.URT.x);
		result->bvol.URT.y = min(geom1->bvol.URT.y, geom2->bvol.URT.y);
		result->bvol.URT.z = min(geom1->bvol.URT.z, geom2->bvol.URT.z);

		result->bvol.LLB.x = max(geom1->bvol.LLB.x, geom2->bvol.LLB.x);
		result->bvol.LLB.y = max(geom1->bvol.LLB.y, geom2->bvol.LLB.y);
		result->bvol.LLB.z = max(geom1->bvol.LLB.z, geom2->bvol.LLB.z);

        if (result->bvol.URT.x < result->bvol.LLB.x || result->bvol.URT.y < result->bvol.LLB.y)
        {
                pfree(result);
                /* Indicate "no intersection" by returning NULL pointer */
                result = NULL;
        }
	  PG_FREE_IF_COPY(geom1, 0);
        PG_FREE_IF_COPY(geom2, 1);

	PG_RETURN_POINTER(result);
}

PG_FUNCTION_INFO_V1(geometry_size);
Datum geometry_size(PG_FUNCTION_ARGS)
{
	 Pointer        	 aptr = PG_GETARG_POINTER(0);

       float      	*size = (float *) PG_GETARG_POINTER(1);
       GEOMETRY    	*a;
	float			xdim,ydim;


//printf("entering rt_points3d_size \n");


        if (aptr == NULL)
        {
                *size = 0.0;
		//printf("	+ aprt==null return in 0.0\n");
                PG_RETURN_VOID();
        }

	        a = (GEOMETRY *)   PG_DETOAST_DATUM(PG_GETARG_DATUM(0));

        if (a->bvol.URT.x <= a->bvol.LLB.x ||
                a->bvol.URT.y <= a->bvol.LLB.y)
                *size = 0.0;
        else
        {
                xdim = (a->bvol.URT.x - a->bvol.LLB.x);
                ydim = (a->bvol.URT.y - a->bvol.LLB.y);

                *size = (float) (xdim * ydim);
        }

		//printf("	+ about to return (and free) with %e\n",*size);

        /* Avoid leaking memory when handed toasted input. */
        PG_FREE_IF_COPY(a, 0);

        PG_RETURN_VOID();
}


/******************************************************
 * GiST index requires these functions
 *
 *  ORGININAL used BOXONLYTYPE geomery as index
 *
 *	NEW version uses geo_decs.h BOX (2d) type
 *
 *	Based on;
 *
 *	Taken from http://www.sai.msu.su/~megera/postgres/gist/ 
 *   {Slightley Modified}
 *
 ******************************************************/

BOX	*convert_box3d_to_box(BOX3D *in)
{
		BOX	*out = palloc (sizeof (BOX) );

		out->high.x = in->URT.x;
		out->high.y = in->URT.y;

		out->low.x = in->LLB.x;
		out->low.y = in->LLB.y;

	return out;
}



GISTENTRY *ggeometry_compress(PG_FUNCTION_ARGS)
{
	GISTENTRY *entry=(GISTENTRY*)PG_GETARG_POINTER(0);
	GISTENTRY *retval;

	if ( entry->leafkey) {
		retval = palloc(sizeof(GISTENTRY));
		if ( entry->pred ) {
			GEOMETRY *in;
			GEOMETRYKEY *r;
			BOX	*thebox;
			in = (GEOMETRY*)PG_DETOAST_DATUM(PointerGetDatum(entry->pred));
			r = (GEOMETRYKEY*)palloc( sizeof(GEOMETRYKEY) );
			r->size = sizeof(GEOMETRYKEY);
			thebox = convert_box3d_to_box(&in->bvol);
			memcpy( (void*)&(r->key), (void*)thebox, sizeof(BOX) );
			if ( (char*)in != entry->pred ) 
			{
				pfree( in );
				pfree(thebox);
			}

			gistentryinit(*retval, (char*)r, entry->rel, entry->page,
				entry->offset, sizeof(GEOMETRYKEY),FALSE);

		} else {
			gistentryinit(*retval, NULL, entry->rel, entry->page,
				entry->offset, 0,FALSE);
		} 
	} else {
		retval = entry;
	}
	return( retval );
}

bool ggeometry_consistent(PG_FUNCTION_ARGS)
{
    GISTENTRY *entry = (GISTENTRY*) PG_GETARG_POINTER(0);
    GEOMETRY *query       = (GEOMETRY*)	    PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
    StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
    BOX		*thebox;
    /*
    ** if entry is not leaf, use gbox_internal_consistent,
    ** else use gbox_leaf_consistent
    */

    if ( ! (entry->pred && query) )
	return FALSE;

	thebox = convert_box3d_to_box( &(query->bvol) );

    PG_RETURN_BOOL(rtree_internal_consistent((BOX*)&( ((GEOMETRYKEY *)(entry->pred))->key ), 
		thebox, strategy));
}


GEOMETRYKEY *ggeometry_union(PG_FUNCTION_ARGS)
{
    return (GEOMETRYKEY*) 
	rtree_union(
		(bytea*) PG_GETARG_POINTER(0),
		(int*) PG_GETARG_POINTER(1),
		ggeometry_binary_union
	);
}


float *ggeometry_penalty(PG_FUNCTION_ARGS)
{
    return rtree_penalty(
	(GISTENTRY*) PG_GETARG_POINTER(0),
	(GISTENTRY*) PG_GETARG_POINTER(1),
	(float*) PG_GETARG_POINTER(2),
	ggeometry_binary_union,
	size_geometrykey
    );
}

GIST_SPLITVEC *ggeometry_picksplit(PG_FUNCTION_ARGS)
{
    return rtree_picksplit(
	(bytea*)PG_GETARG_POINTER(0),
	(GIST_SPLITVEC*)PG_GETARG_POINTER(1),
	sizeof(GEOMETRYKEY),
	ggeometry_binary_union,
	ggeometry_inter,
	size_geometrykey
    );
}

bool *ggeometry_same(PG_FUNCTION_ARGS)
{
  GEOMETRYKEY *b1 = (GEOMETRYKEY*) PG_GETARG_POINTER(0);
  GEOMETRYKEY *b2 = (GEOMETRYKEY*) PG_GETARG_POINTER(1);

  bool *result = (bool*) PG_GETARG_POINTER(2);
  if ( b1 && b2 )
   	*result = DatumGetBool( DirectFunctionCall2( box_same, 
		PointerGetDatum(&(b1->key)), 
		PointerGetDatum(&(b2->key))) );
  else
	*result = ( b1==NULL && b2==NULL ) ? TRUE : FALSE; 
  return(result);
}

Datum ggeometry_inter(PG_FUNCTION_ARGS) {
  	GEOMETRYKEY *b1 = (GEOMETRYKEY*) PG_GETARG_POINTER(0);
  	GEOMETRYKEY*b2 = (GEOMETRYKEY*) PG_GETARG_POINTER(1);
	char *interd;

    	interd = DatumGetPointer(DirectFunctionCall2(
			rt_box_inter,
			PointerGetDatum( &(b1->key) ),
			PointerGetDatum( &(b2->key) )) );
	
	if (interd) {
		GEOMETRYKEY *tmp = (GEOMETRYKEY*)palloc( sizeof(GEOMETRYKEY) );
		tmp->size = sizeof(GEOMETRYKEY);
		memcpy( (void*)&(tmp->key), (void*)interd, sizeof(BOX) );
		pfree( interd );
		PG_RETURN_POINTER( tmp );
	} else 
		PG_RETURN_POINTER( NULL );
}

static char *ggeometry_binary_union(char *r1, char *r2, int *sizep)
{
    GEOMETRYKEY *retval;

    if ( ! (r1 && r2) ) {
	if ( r1 ) {
		retval = (GEOMETRYKEY*)palloc( sizeof(GEOMETRYKEY) );
		memcpy( (void*)retval, (void*)r1, sizeof(GEOMETRYKEY) );
    		*sizep = sizeof(GEOMETRYKEY);
	} else if ( r2 ) {
		retval = (GEOMETRYKEY*)palloc( sizeof(GEOMETRYKEY) );
		memcpy( (void*)retval, (void*)r2, sizeof(GEOMETRYKEY) );
    		*sizep = sizeof(GEOMETRYKEY);
	} else {
		*sizep = 0;
		retval = NULL;
	} 
    } else {
    	BOX *key = (BOX*)DatumGetPointer( DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( &(((GEOMETRYKEY*)r1)->key) ),
		PointerGetDatum( &(((GEOMETRYKEY*)r2)->key) )) );
	retval = (GEOMETRYKEY*)palloc( sizeof(GEOMETRYKEY) );
	memcpy( (void*)&(retval->key), (void*)key, sizeof(BOX) );
	pfree( key );
    	*sizep = retval->size = sizeof(GEOMETRYKEY);
    }
    return (char*)retval;
}


static float size_geometrykey( char *pk ) {
    if ( pk ) {
	float size;
    	DirectFunctionCall2( rt_box_size,
		PointerGetDatum( &(((GEOMETRYKEY*)pk)->key) ),
		PointerGetDatum( &size ) );
	return size;
    } else
	return 0.0;
}

static char *rtree_union(bytea *entryvec, int *sizep, BINARY_UNION bu)
{
    int numranges, i;
    char *out, *tmp;

    numranges = (VARSIZE(entryvec) - VARHDRSZ)/sizeof(GISTENTRY); 
    tmp = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[0]).pred;
    out = NULL;

    for (i = 1; i < numranges; i++) {
	out = (*bu)(tmp, (char *)
				 (((GISTENTRY *)(VARDATA(entryvec)))[i]).pred,
				 sizep);
	if (i > 1 && tmp) pfree(tmp);
	tmp = out;
    }

    return(out);
}

static float *rtree_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result, BINARY_UNION bu, SIZE_BOX sb)
{
    char * ud;
    float tmp1;
    int sizep;
   
    ud = (*bu)( origentry->pred, newentry->pred, &sizep );
    tmp1 = (*sb)( ud ); 
    if (ud) pfree(ud);

    *result = tmp1 - (*sb)( origentry->pred );
    return(result);
}

/*
** The GiST PickSplit method
** We use Guttman's poly time split algorithm 
*/
static GIST_SPLITVEC *rtree_picksplit(bytea *entryvec, GIST_SPLITVEC *v, int keylen, BINARY_UNION bu, RDF interop, SIZE_BOX sb)
{
    OffsetNumber i, j;
    char *datum_alpha, *datum_beta;
    char *datum_l, *datum_r;
    char *union_d, *union_dl, *union_dr;
    char *inter_d;
    bool firsttime;
    float size_alpha, size_beta, size_union, size_inter;
    float size_waste, waste;
    float size_l, size_r;
    int nbytes;
    int sizep;
    OffsetNumber seed_1 = 0, seed_2 = 0;
    OffsetNumber *left, *right;
    OffsetNumber maxoff;

    maxoff = ((VARSIZE(entryvec) - VARHDRSZ)/sizeof(GISTENTRY)) - 2;
    nbytes =  (maxoff + 2) * sizeof(OffsetNumber);
    v->spl_left = (OffsetNumber *) palloc(nbytes);
    v->spl_right = (OffsetNumber *) palloc(nbytes);
    
    firsttime = true;
    waste = 0.0;
    
    for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i)) {
	datum_alpha = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[i].pred);
	for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j)) {
	    datum_beta = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[j].pred);
	    
	    /* compute the wasted space by unioning these guys */
	    /* size_waste = size_union - size_inter; */	
	    union_d = (*bu)( datum_alpha, datum_beta, &sizep );
	    if ( union_d ) {
		size_union = (*sb)(union_d);
		pfree(union_d);
	    } else
		size_union = 0.0;

	    if ( datum_alpha && datum_beta ) {
    	    	inter_d = DatumGetPointer(DirectFunctionCall2(
			interop,
			PointerGetDatum( datum_alpha ),
			PointerGetDatum( datum_beta )) );
		if ( inter_d ) {
			size_inter = (*sb)(inter_d);
			pfree(inter_d);
		} else 
			size_inter = 0.0;
	    } else
		size_inter = 0.0;

	    size_waste = size_union - size_inter;
	    
	    /*
	     *  are these a more promising split that what we've
	     *  already seen?
	     */
	    
	    if (size_waste > waste || firsttime) {
		waste = size_waste;
		seed_1 = i;
		seed_2 = j;
		firsttime = false;
	    }
	}
    }
    
    left = v->spl_left;
    v->spl_nleft = 0;
    right = v->spl_right;
    v->spl_nright = 0;
  
    if ( ((GISTENTRY *)(VARDATA(entryvec)))[seed_1].pred ) {
	datum_l = (char*) palloc( keylen );
	memcpy( (void*)datum_l, (void*)(((GISTENTRY *)(VARDATA(entryvec)))[seed_1].pred ), keylen );
    } else 
	datum_l = NULL; 
    size_l  = (*sb)( datum_l ); 
    if ( ((GISTENTRY *)(VARDATA(entryvec)))[seed_2].pred ) {
	datum_r = (char*) palloc( keylen );
	memcpy( (void*)datum_r, (void*)(((GISTENTRY *)(VARDATA(entryvec)))[seed_2].pred ), keylen );
    } else 
	datum_r = NULL; 
    size_r  = (*sb)( datum_r ); 
    
    /*
     *  Now split up the regions between the two seeds.  An important
     *  property of this split algorithm is that the split vector v
     *  has the indices of items to be split in order in its left and
     *  right vectors.  We exploit this property by doing a merge in
     *  the code that actually splits the page.
     *
     *  For efficiency, we also place the new index tuple in this loop.
     *  This is handled at the very end, when we have placed all the
     *  existing tuples and i == maxoff + 1.
     */
    
    maxoff = OffsetNumberNext(maxoff);
    for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
	
	/*
	 *  If we've already decided where to place this item, just
	 *  put it on the right list.  Otherwise, we need to figure
	 *  out which page needs the least enlargement in order to
	 *  store the item.
	 */
	
	if (i == seed_1) {
	    *left++ = i;
	    v->spl_nleft++;
	    continue;
	} else if (i == seed_2) {
	    *right++ = i;
	    v->spl_nright++;
	    continue;
	}
	
	/* okay, which page needs least enlargement? */ 
	datum_alpha = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[i].pred);
	union_dl = (*bu)( datum_l, datum_alpha, &sizep );
	union_dr = (*bu)( datum_r, datum_alpha, &sizep );
	size_alpha = (*sb)( union_dl ); 
	size_beta  = (*sb)( union_dr ); 
	
	/* pick which page to add it to */
	if (size_alpha - size_l < size_beta - size_r) {
	    pfree(datum_l);
	    pfree(union_dr);
	    datum_l = union_dl;
	    size_l = size_alpha;
	    *left++ = i;
	    v->spl_nleft++;
	} else {
	    pfree(datum_r);
	    pfree(union_dl);
	    datum_r = union_dr;
	    size_r = size_alpha;
	    *right++ = i;
	    v->spl_nright++;
	}
    }
    *left = *right = FirstOffsetNumber;	/* sentinel value, see dosplit() */
    
    v->spl_ldatum = datum_l;
    v->spl_rdatum = datum_r;

    return( v );
}


static bool rtree_internal_consistent(BOX *key,
			BOX *query,
			StrategyNumber strategy)
{
    bool retval;

    switch(strategy) {
    case RTLeftStrategyNumber:
    case RTOverLeftStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overleft, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTOverlapStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overlap, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTOverRightStrategyNumber:
    case RTRightStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_right, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTSameStrategyNumber:
    case RTContainsStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_contain, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTContainedByStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overlap, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    default:
      retval = FALSE;
    }
    return(retval);
}

GISTENTRY *rtree_decompress(PG_FUNCTION_ARGS)
{
    return((GISTENTRY*)PG_GETARG_POINTER(0));
}


/**************************************************************************
 * GENERAL PURPOSE GEOMETRY FUNCTIONS
 **************************************************************************/


double line_length2d(LINE3D *line)
{
	int	i;
	POINT3D	*frm,*to;
	double	dist = 0.0;

	if (line->npoints <2)
		return 0.0;	//must have >1 point to make sense

	frm = &line->points[0];

	for (i=1; i<line->npoints;i++)
	{
		to = &line->points[i];
	
		dist += sqrt( ( (frm->x - to->x)*(frm->x - to->x) )  +
			  	  ( (frm->y - to->y)*(frm->y - to->y) ) );

		frm = to;
	}
	return dist;
}

double line_length3d(LINE3D *line)
{
	int	i;
	POINT3D	*frm,*to;
	double	dist = 0.0;

	if (line->npoints <2)
		return 0.0;	//must have >1 point to make sense

	frm = &line->points[0];

	for (i=1; i<line->npoints;i++)
	{
		to = &line->points[i];
	
		dist += sqrt( ( (frm->x - to->x)*(frm->x - to->x) )  +
				((frm->y - to->y)*(frm->y - to->y) ) + 
				((frm->z - to->z)*(frm->z - to->z) ) );

		frm = to;
	}
	return dist;
}


//find the "length of a geometry"
// length3d(point) = 0
// length3d(line) = length of line
// length3d(polygon) = 0  -- could make sense to return sum(ring perimeter)
// uses euclidian 3d length
 
PG_FUNCTION_INFO_V1(length3d);
Datum length3d(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int32				*offsets1;
	char				*o1;
	int32				type1,j;
	LINE3D			*line;
	double			dist = 0.0;

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;


	//now have to do a scan of each object

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];
		if (type1 == LINETYPE)	//LINESTRING
		{
			line = (LINE3D *) o1;
			dist += line_length3d(line);
		}
	}
	PG_RETURN_FLOAT4(dist);
}

//find the "length of a geometry"
// length3d(point) = 0
// length3d(line) = length of line
// length3d(polygon) = 0  -- could make sense to return sum(ring perimeter)
// uses euclidian 2d length
 
PG_FUNCTION_INFO_V1(length2d);
Datum length2d(PG_FUNCTION_ARGS)
{
	GEOMETRY		   *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int32				*offsets1;
	char				*o1;
	int32				type1,j;
	LINE3D			*line;
	double			dist = 0.0;

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;


	//now have to do a scan of each object

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];
		if (type1 == LINETYPE)	//LINESTRING
		{
			line = (LINE3D *) o1;
			dist += line_length2d(line);
		}
	}
	PG_RETURN_FLOAT4(dist);
}


//find the 2d area of the outer ring - sum (area 2d of inner rings)
// Could use a more numerically stable calculator...
double polygon_area2d_old(POLYGON3D *poly1)
{
	double	poly_area=0.0, ringarea=0.0;
	int i,j,ring,pt_offset;
	POINT3D	*pts1;


	pts1 = (POINT3D *) ( (char *)&(poly1->npoints[poly1->nrings] )  );
	pts1 = (POINT3D *) MAXALIGN(pts1);


	
	pt_offset = 0; //index to first point in ring
	for (ring = 0; ring < poly1->nrings; ring++)
	{
		ringarea = 0.0;
		
		for (i=0;i<(poly1->npoints[ring]);i++) 
    		{ 
			j = (i+1) % (poly1->npoints[ring]);
			ringarea  += pts1[pt_offset+ i].x * pts1[pt_offset+j].y - pts1[pt_offset+ i].y * pts1[pt_offset+j].x; 
		} 

		ringarea  /= 2.0; 

		ringarea  = fabs(ringarea );
		if (ring != 0)	//outer
			ringarea  = -1.0*ringarea ; // its a hole
		
		poly_area += ringarea;

		pt_offset += poly1->npoints[ring];	
	}

	return poly_area;
}



//calculate the area of all the subobj in a polygon
// area(point) = 0
// area (line) = 0
// area(polygon) = find its 2d area
PG_FUNCTION_INFO_V1(area2d);
Datum area2d(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int32				*offsets1;
	char				*o1;
	int32				type1,j;
	POLYGON3D			*poly;
	double			area = 0.0;

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;

	//now have to do a scan of each object

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];
		if (type1 == POLYGONTYPE)	//POLYGON
		{
			poly = (POLYGON3D *) o1;
			area += polygon_area2d_old(poly);
		}
	}
	PG_RETURN_FLOAT4(area);
}

double 	polygon_perimeter3d(POLYGON3D	*poly1)
{
	double	poly_perimeter=0.0, ring_perimeter=0.0;
	int i,ring,pt_offset;
	POINT3D	*pts1;
	POINT3D	*to,*frm;


	pts1 = (POINT3D *) ( (char *)&(poly1->npoints[poly1->nrings] )  );
	pts1 = (POINT3D *) MAXALIGN(pts1);
	

	
	pt_offset = 0; //index to first point in ring
	for (ring = 0; ring < poly1->nrings; ring++)
	{
		ring_perimeter = 0.0;
		

		frm = &pts1[pt_offset];

		for (i=1; i<poly1->npoints[ring];i++)
		{
			to = &pts1[pt_offset+ i];
	
			ring_perimeter += sqrt( ( (frm->x - to->x)*(frm->x - to->x) )  +
				((frm->y - to->y)*(frm->y - to->y) ) + 
				((frm->z - to->z)*(frm->z - to->z) ) );

			frm = to;
		}
		
		poly_perimeter += ring_perimeter;

		pt_offset += poly1->npoints[ring];	
	}

	return poly_perimeter;

}

double 	polygon_perimeter2d(POLYGON3D	*poly1)
{
	double	poly_perimeter=0.0, ring_perimeter=0.0;
	int i,ring,pt_offset;
	POINT3D	*pts1;
	POINT3D	*to,*frm;


	pts1 = (POINT3D *) ( (char *)&(poly1->npoints[poly1->nrings] )  );
	pts1 = (POINT3D *) MAXALIGN(pts1);

	
	pt_offset = 0; //index to first point in ring
	for (ring = 0; ring < poly1->nrings; ring++)
	{
		ring_perimeter = 0.0;
		

		frm = &pts1[pt_offset];

		for (i=1; i<poly1->npoints[ring];i++)
		{
			to = &pts1[pt_offset+ i];
	
			ring_perimeter += sqrt( ( (frm->x - to->x)*(frm->x - to->x) )  +
				((frm->y - to->y)*(frm->y - to->y) )   );

			frm = to;
		}
		
		poly_perimeter += ring_perimeter;

		pt_offset += poly1->npoints[ring];	
	}

	return poly_perimeter;

}



//calculate the perimeter of polys (sum of length of all rings)
PG_FUNCTION_INFO_V1(perimeter3d);
Datum perimeter3d(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int32				*offsets1;
	char				*o1;
	int32				type1,j;
	POLYGON3D			*poly;
	double			area = 0.0;

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;

	//now have to do a scan of each object

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];
		if (type1 == POLYGONTYPE)	//POLYGON
		{
			poly = (POLYGON3D *) o1;
			area += polygon_perimeter3d(poly);
		}
	}
	PG_RETURN_FLOAT4(area);
}

//calculate the perimeter of polys (sum of length of all rings)
PG_FUNCTION_INFO_V1(perimeter2d);
Datum perimeter2d(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int32				*offsets1;
	char				*o1;
	int32				type1,j;
	POLYGON3D			*poly;
	double			area = 0.0;

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;

	//now have to do a scan of each object

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];
		if (type1 == POLYGONTYPE)	//POLYGON
		{
			poly = (POLYGON3D *) o1;
			area += polygon_perimeter2d(poly);
		}
	}
	PG_RETURN_FLOAT4(area);
}


// cn_PnPoly(): crossing number test for a point in a polygon
//      input:   P = a point,
//               V[] = vertex points of a polygon V[n+1] with V[n]=V[0]
//      returns: 0 = outside, 1 = inside
//
//	Our polygons have first and last point the same, 
//

int PIP( POINT3D *P, POINT3D *V, int n )
{
    int    cn = 0;    // the crossing number counter
 		int i;

	double vt;

    // loop through all edges of the polygon

    	for (i=0; i< (n-1) ; i++) 
	{    // edge from V[i] to V[i+1]

       	if (((V[i].y <= P->y) && (V[i+1].y > P->y))    // an upward crossing
       		 || ((V[i].y > P->y) && (V[i+1].y <= P->y))) // a downward crossing
	 	{ 

       	   vt = (double)(P->y - V[i].y) / (V[i+1].y - V[i].y);
         	   if (P->x < V[i].x + vt * (V[i+1].x - V[i].x)) 			// P.x <intersect
           	     		++cn;   // a valid crossing of y=P.y right of P.x
      	}	
    }
    return (cn&1);    // 0 if even (out), and 1 if odd (in)

}





//this one is easy - is the point inside a box?
bool point_truely_inside(POINT3D	*point, BOX3D	*box)
{
	return ( 
			(point->x >= box->LLB.x) && (point->x <= box->URT.x)  && 
		      (point->y >= box->LLB.y) && (point->y <= box->URT.y)  
		 );
}


// see if a linestring is inside a box - check to see if each its line
//	segments are inside the box  
// NOTE: all this is happening in 2d - third dimention is ignored
//
//use a modified CohenSutherland line clipping algoritm
//
//Step one - compute outcodes for Pi and Pi+1
//
//
// 			|		|
//		1001	|   1000	|  1010
//			|		|
//      ----------+-----------+------------
// 			| (inside)	|
//		0001	|  		|  0010
//			|   0000	|
//      ----------+-----------+------------
// 			|		|
//		0101	|   0100	|  0110
//			|		|
//
// If Pi or Pi+1 have code 0000 then this line seg is inside the box  (trivial accept)
//  if Pi && Pi+1 != 0 then seg is NOT in the box (trivial reject)
//
//	Otherwise the line seg might traverse through the box.
//		=> see if the seg intersects any of the walls of the box.
//

int	compute_outcode( POINT3D *p, BOX3D *box)
{
	int	Code=0;

//printf("compute outcode:\n");
//print_box(box);
//printf("point = [%g,%g]\n",p->x,p->y);
	if (p->y > box->URT.y )
		Code = 8;
  	else
  	{
	  if (p->y <box->LLB.y )
	  {
		Code = 4;
	  }
  	}
  	if (p->x > box->URT.x )
  	{
	  return (Code+2);
 	}
 	if (p->x <box->LLB.x )
 	{
	  return (Code+1);
 	}

 	return (Code);		
}


bool lineseg_inside_box( POINT3D *P1, POINT3D *P2, BOX3D *box)
{
	int	outcode_p1, outcode_p2;
	double	Ax,Ay, Bx,By, Cx,Cy, Dx,Dy;
	double	r,s;

	outcode_p1 = compute_outcode(P1, box);
	if (outcode_p1 ==0)
		return TRUE;
	
	outcode_p2 = compute_outcode(P2, box);
	if (outcode_p2 ==0)
		return TRUE;
	if ((outcode_p1 & outcode_p2) != 0)
		return FALSE;


	if ((outcode_p1 + outcode_p2) == 12)
		return TRUE; //vertically slices box

	if ((outcode_p1 + outcode_p2) == 3)
		return TRUE; //horizontally slices box


//printf("have to do boundry calcs\n");

	// know that its a diagonal line

	//this is the tough case - it may or maynot intersect the box
	//see if it intersects the horizonal box lines

	// from comp.graphics.algo's faq:
	/*
	       (Ay-Cy)(Dx-Cx)-(Ax-Cx)(Dy-Cy) 
	r =  ------------------------------------
         	  (Bx-Ax)(Dy-Cy)-(By-Ay)(Dx-Cx) 

		(Ay-Cy)(Bx-Ax)-(Ax-Cx)(By-Ay) 
	s = 	-----------------------------------
		(Bx-Ax)(Dy-Cy)-(By-Ay)(Dx-Cx) 

	*/
	// if  0<= r&s <=1 then intersection exists
	
	Ax = P1->x; Ay = P1->y;
	Bx = P2->x; By = P2->y;

	//top hor part of box
	Cx = box->LLB.x; Cy= box->URT.y;
	Dx = box->URT.x; Dy= box->URT.y;

		r= ((Ay-Cy)*(Dx-Cx)-(Ax-Cx)*(Dy-Cy) )/
			((Bx-Ax)*(Dy-Cy)-(By-Ay)*(Dx-Cx) ) ;

		s=  ((Ay-Cy)*(Bx-Ax)-(Ax-Cx)*(By-Ay))/
			((Bx-Ax)*(Dy-Cy)-(By-Ay)*(Dx-Cx) ) ;

	if  ( (r>=0) && (r<=1) && (s>=0) && (s<=1)    )
		return TRUE;

	//bottom hor part of box
	Cx = box->LLB.x; Cy= box->LLB.y;
	Dx = box->URT.x; Dy= box->LLB.y;

		r= ((Ay-Cy)*(Dx-Cx)-(Ax-Cx)*(Dy-Cy) )/
			((Bx-Ax)*(Dy-Cy)-(By-Ay)*(Dx-Cx) ) ;

		s=  ((Ay-Cy)*(Bx-Ax)-(Ax-Cx)*(By-Ay))/
			((Bx-Ax)*(Dy-Cy)-(By-Ay)*(Dx-Cx) ) ;
	if  ( (r>=0) && (r<=1) && (s>=0) && (s<=1)    )
		return TRUE;

	return FALSE;
	
}


//give the work to an easier process
bool	linestring_inside_box(POINT3D *pts, int npoints, BOX3D *box)
{
	POINT3D	*frm,*to;
	int		i;

	if (npoints <2)
		return FALSE;	//must have >1 point to make sense

	frm = &pts[0];

	for (i=1; i< npoints;i++)
	{
		to = &pts[i];
	
		if (lineseg_inside_box( frm, to,  box)) 
			return TRUE;

		frm = to;
	}
	return FALSE;
}


//this ones the most complex
//	1. treat the outer ring as a line and see if its inside
//		+ if it is, then poly is inside box
//	2. The polygon could be completely contained inside the box
//		+ line detector (step 1) would have found this
//	3. The polygon could be completely surrounding the box
//		+ point-in-polygon of one of the box corners
//		+ COMPLEXITY: box could be totally inside a single hole
//			+test all holes like step 1.  Any that pass arent "bad" holes
//			+  each of the bad holes do a point-in-polygon if true =-> totally in hole

bool polygon_truely_inside(POLYGON3D	*poly, BOX3D *box)
{
	bool		outer_ring_inside,outer_ring_surrounds;
	POINT3D	*pts1;
	POINT3D	test_point;
	int32		ring;
	int		point_offset;
	
	pts1 = (POINT3D *) ( (char *)&(poly->npoints[poly->nrings] )  );
	pts1 = (POINT3D *) MAXALIGN(pts1);


	//step 1 (and 2)
	outer_ring_inside = linestring_inside_box(pts1, poly->npoints[0], box);	

	if (outer_ring_inside)
		return TRUE;

	test_point.x = box->LLB.x;
	test_point.y = box->LLB.y;
		//.z unimportant
	
	outer_ring_surrounds = PIP(&test_point, pts1,poly->npoints[0] ); 

	if (!(outer_ring_surrounds))
		return FALSE;	// disjoing

	//printf("polygon - have to check holes\n");

	//step 3

	point_offset = poly->npoints[0];
	for (ring =1; ring< poly->nrings; ring ++)
	{
		if (!( linestring_inside_box(&pts1[point_offset], poly->npoints[ring], box) ))
		{
			//doesnt intersect the box
			// so if one of the corners of the box is inside the hole, it totally
			//   surrounds the box

			if (PIP(&test_point, &pts1[point_offset],poly->npoints[ring] ) )
				return FALSE;
		}
		point_offset += poly->npoints[ring];
	}
	
	return TRUE;
}





//This is more complicated because all the points in the line can be outside the box,
// but one of the edges can cross into the box.
//  We send the work off to another function because its generically usefull for
//   polygons too
bool line_truely_inside( LINE3D *line, BOX3D 	*box)
{
	return	linestring_inside_box(&line->points[0], line->npoints, box);
}

//is anything in geom1 really inside the the bvol (2d only) defined in geom2?
// send each object to its proper handler
PG_FUNCTION_INFO_V1(truly_inside);
Datum truly_inside(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY		      *geom2 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
	BOX3D				*the_bbox;
	int32				*offsets1;
	char				*o1;
	int32				type1,j;
	POLYGON3D			*poly;
	LINE3D			*line;
	POINT3D			*point;

	the_bbox = &geom2->bvol;

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;

	//now have to do a scan of each object 

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];

		if (type1 == POINTTYPE)	//point
		{
			point = (POINT3D *) o1;
			if ( point_truely_inside(point, the_bbox))
				PG_RETURN_BOOL(TRUE);
		}

		if (type1 == LINETYPE)	//line
		{
			line = (LINE3D *) o1;
			if ( line_truely_inside(line, the_bbox))
				PG_RETURN_BOOL(TRUE);
		}

		if (type1 == POLYGONTYPE)	//POLYGON
		{
			poly = (POLYGON3D *) o1;

			if ( polygon_truely_inside(poly, the_bbox))
				PG_RETURN_BOOL(TRUE);
		}
	}
	PG_RETURN_BOOL(FALSE);
}



//number of points in an object
PG_FUNCTION_INFO_V1(npoints);
Datum npoints(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int32				*offsets1;
	char				*o1;
	int32				type1,j,i;
	POLYGON3D			*poly;
	LINE3D			*line;
	int32				numb_points = 0;

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;

	//now have to do a scan of each object 

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];

		if (type1 == POINTTYPE)	//point
		{
			numb_points ++;
		}

		if (type1 == LINETYPE)	//line
		{
			line = (LINE3D *) o1;
			numb_points += line->npoints;
		}

		if (type1 == POLYGONTYPE)	//POLYGON
		{
			poly = (POLYGON3D *) o1;
			for (i=0; i<poly->nrings;i++)
			{
				numb_points += poly->npoints[i];
			}
		}
	}

	PG_RETURN_INT32(numb_points);
}

//number of rings in an object
PG_FUNCTION_INFO_V1(nrings);
Datum nrings(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int32				*offsets1;
	char				*o1;
	int32				type1,j;
	POLYGON3D			*poly;
	int32				numb_rings = 0;

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;

	//now have to do a scan of each object 

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];

		if (type1 == POLYGONTYPE)	//POLYGON
		{
			poly = (POLYGON3D *) o1;
			numb_rings += poly->nrings;		
		}
	}

	PG_RETURN_INT32(numb_rings);
}


//find the size of geometry
PG_FUNCTION_INFO_V1(mem_size);
Datum mem_size(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));

	PG_RETURN_INT32(geom1->size);
}

//find the size of geometry
PG_FUNCTION_INFO_V1(numb_sub_objs);
Datum numb_sub_objs(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));

	PG_RETURN_INT32(geom1->nobjs);
}

//get summary info on a GEOMETRY
PG_FUNCTION_INFO_V1(summary);
Datum summary(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom1 = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	int32				*offsets1;
	char				*o1;
	int32				type1,j,i;
	POLYGON3D			*poly;
	LINE3D			*line;
	char				*result;
	int				size;
	char				tmp[100];
	text				*mytext;


	size = 1;
	result = palloc(1);
	result[0] = 0; //null terminate it
	

	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;

	//now have to do a scan of each object 

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type1=  geom1->objType[j];

		if (type1 == POINTTYPE)	//point
		{
			size += 30;
			result = repalloc(result,size);
			sprintf(tmp,"Object %i is a POINT()\n",j);
			strcat(result,tmp);
		}

		if (type1 == LINETYPE)	//line
		{
			line = (LINE3D *) o1;

			size += 57;
			result = repalloc(result,size);
			sprintf(tmp,"Object %i is a LINESTRING() with %i points\n",j,line->npoints);
			strcat(result,tmp);
		}

		if (type1 == POLYGONTYPE)	//POLYGON
		{
			poly = (POLYGON3D *) o1;
		
			size += 57*(poly->nrings +1);
			result = repalloc(result,size);
			sprintf(tmp,"Object %i is a POLYGON() with %i rings\n",j,poly->nrings);
			strcat(result,tmp);
			for (i=0; i<poly->nrings;i++)
			{	
				sprintf(tmp,"     + ring %i has %i points\n",i,poly->npoints[i]);
				strcat(result,tmp);
			}
		}
	}

		// create a text obj to return
	mytext = (text *) palloc(VARHDRSZ  + strlen(result) );
	VARATT_SIZEP(mytext) = VARHDRSZ + strlen(result) ;
	memcpy(VARDATA(mytext) , result, strlen(result) );
	pfree(result);
	PG_RETURN_POINTER(mytext);
}

void	translate_points(POINT3D *pt, int npoints,double x_off, double y_off, double z_off)
{
	int i;

//printf("translating %i points by [%g,%g,%g]\n",npoints,x_off,y_off,z_off);
	if (npoints <1)
		return; //nothing to do

	for (i=0;i<npoints;i++)
	{
//printf("before: [%g,%g,%g]\n", pt[i].x, pt[i].y,pt[i].z);

		pt[i].x += x_off;
		pt[i].y += y_off;
		pt[i].z += z_off;

//printf("after: [%g,%g,%g]\n", pt[i].x, pt[i].y,pt[i].z);
	}
}



//translate geometry
PG_FUNCTION_INFO_V1(translate);
Datum translate(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
	GEOMETRY			*geom1;
	double			x_off =  PG_GETARG_FLOAT8(1);
	double			y_off =  PG_GETARG_FLOAT8(2);
	double			z_off =  PG_GETARG_FLOAT8(3);
	int32				*offsets1;
	char				*o1;
	int32				type,j,i;
	POLYGON3D			*poly;
	POINT3D			*point,*pts;
	LINE3D			*line;
	int				numb_points;




	//make a copy of geom so we can return a new version

	geom1 = (GEOMETRY *) palloc (geom->size);
	memcpy(geom1, geom, geom->size);


	offsets1 = (int32 *) ( ((char *) &(geom1->objType[0] ))+ sizeof(int32) * geom1->nobjs ) ;

	//now have to do a scan of each object 

	for (j=0; j< geom1->nobjs; j++)		//for each object in geom1
	{
		o1 = (char *) geom1 +offsets1[j] ;  
		type=  geom1->objType[j];

		if (type == POINTTYPE)	//point
		{
			point = (POINT3D *) o1;
			translate_points(point, 1,x_off,y_off,z_off);
		}

		if (type == LINETYPE)	//line
		{
			line = (LINE3D *) o1;
			translate_points(&(line->points[0]), line->npoints,x_off,y_off,z_off);
		}

		if (type == POLYGONTYPE)	//POLYGON
		{
			poly = (POLYGON3D *) o1;
			//find where the points are and where to put them
			numb_points =0;
			for (i=0; i<poly->nrings;i++)
			{
				numb_points += poly->npoints[i];
			}
			pts = (POINT3D *) ( (char *)&(poly->npoints[poly->nrings] )  );
			pts = (POINT3D *) MAXALIGN(pts);
			translate_points(pts, numb_points,x_off,y_off,z_off);


		}
	}
	PG_RETURN_POINTER(geom1);
}

//*****************************************************************
//WKB does not do any padding or alignment!

// in general these function take 
//	1. geometry item or list
//	2. if the endian-ness should be changed (flipbytes)
//	3. byte_order - 0=big endian (XDR) 1= little endian (NDR)
//    4. use the 3d extention wkb (use3d)
//
// NOTE:
//		A. mem - if null, function pallocs memory for result
//			 - otherwise, use mem to dump the wkb into
//		b. for lists, the # of sub-objects is also specified
//
//  and return
//	1. size of the wkb structure ("int *size")
//	2. the actual wkb structure
//*****************************************************************

//convert a POINT3D into wkb
char	*wkb_point(POINT3D *pt,int32 *size, bool flipbytes, char byte_order, bool use3d)
{
	char		*result ;
	uint32	type ; 
	char		*c_type = (char *) &type;

	if (use3d)
	{
		*size = 29; 
		type = 32768 + 1;	
	}
	else
	{
		*size = 21; 
		type =  1;	
	}

	result = palloc (*size);

	if (flipbytes)
		flip_endian_int32( (char *) &type);

	result[0] = byte_order;
	result[1] = c_type[0];
	result[2] = c_type[1];
	result[3] = c_type[2];
	result[4] = c_type[3];

	if (use3d)
		memcpy(&result[5], pt, 24 ); //copy in point data
	else
		memcpy(&result[5], pt, 16 ); //copy in point data

	if (flipbytes)
	{
		flip_endian_double((char *) &result[5]);		// x
		flip_endian_double((char *) &result[5+8]);	// y
		if (use3d)
			flip_endian_double((char *) &result[5+16]);	// z
	}

	return result;
}

//pt is a pointer to list of POINT3Ds
char	*wkb_multipoint(POINT3D *pt,int32 numb_points,int32 *size, bool flipbytes, char byte_order,bool use3d)
{
	char *result;
	uint32	type ; 		// 3D/2D multipoint, following frank's spec
	uint32	npoints = numb_points;
	char		*c_type = (char *) &type;
	char		*c_npoints = (char *) &npoints;
	int		t;
	int32		junk;
	const int		sub_size_3d = 29;
	const int		sub_size_2d = 21;
	



	if (use3d)
	{
		*size = 9 + sub_size_3d*numb_points; 
		type = 32768 + 4;	
	}
	else
	{
		*size = 9 + sub_size_2d*numb_points; 
		type =  4;	
	}


	result = palloc (*size);
	npoints = numb_points;

	if (flipbytes)
	{
		flip_endian_int32((char *) &type);
		flip_endian_int32((char *) &npoints);
	}


	result[0] = byte_order;
	result[1] = c_type[0];
	result[2] = c_type[1];
	result[3] = c_type[2];
	result[4] = c_type[3];

	result[5] = c_npoints[0];
	result[6] = c_npoints[1];
	result[7] = c_npoints[2];
	result[8] = c_npoints[3];

	for (t=0; t<numb_points;t++)  //for each point
	{
		if (use3d)
			memcpy(&result[9+ t*sub_size_3d], 
					wkb_point(&pt[t],&junk,  flipbytes,  byte_order,use3d)    
						, sub_size_3d);
		else
			memcpy(&result[9+ t*sub_size_2d], 
					wkb_point(&pt[t],&junk,  flipbytes,  byte_order,use3d) 
							, sub_size_2d);	
	}

	return result;

}


//if mem != null, it must have enough space for the wkb
char	*wkb_line(LINE3D *line,int32 *size, bool flipbytes, char byte_order,bool use3d, char *mem)
{
	char *result;
	uint32	type ; 	
	
	char		*c_type = (char *) &type;
	int		t;

	uint32	npoints = line->npoints;
	int		numb_points = line->npoints;
	char		*c_npoints = (char *) &npoints;


	if (use3d)
	{
		*size = 9 + 24*numb_points; 
		type = 32768 + 2;	
	}
	else
	{
		*size = 9 + 16*numb_points; 
		type =  2;	
	}



	if (flipbytes)
	{
		flip_endian_int32((char *) &type);
		flip_endian_int32((char *) &npoints);
	}

	if (mem == NULL)
		result = palloc (*size);
	else
		result = mem;

	result[0] = byte_order;	
	result[1] = c_type[0];		//type
	result[2] = c_type[1];
	result[3] = c_type[2];
	result[4] = c_type[3];

	result[5] = c_npoints[0];	//npoints
	result[6] = c_npoints[1];
	result[7] = c_npoints[2];
	result[8] = c_npoints[3];

	if (use3d)
	{
		memcpy(&result[9],&line->points, line->npoints*24); // copy all the pt data in one chuck
		if (flipbytes)
		{
			for (t=0;t<line->npoints; t++)
			{
				flip_endian_double((char *) &result[9   + t*24]);	// x
				flip_endian_double((char *) &result[9+8 + t*24]);	// y
				flip_endian_double((char *) &result[9+16+ t*24]);	// z
			}
		}
	}
	else
	{
		for (t=0; t<numb_points;t++)
		{
			memcpy(&result[9+16*t],&((&line->points)[t]), 16);  //need to copy each pt individually
			if (flipbytes)
			{
				flip_endian_double((char *) &result[9   + t*16]);	// x
				flip_endian_double((char *) &result[9+8 + t*16]);	// y
			}
		}
	}
	return result;
}

// expects an array of pointers to lines
// calls wkb_line() to do most of the work
char	*wkb_multiline(LINE3D **lines,int32 *size, int numb_lines, bool flipbytes, char byte_order,bool use3d)
{
	int	total_points = 0;
	int	total_size=0;
	int	size_part = 0;

	int 	t;
	char	*result,*new_spot;


	uint32	type ; 	
	char		*c_type = (char *) &type;

	uint32	nlines = numb_lines;
	char		*c_nlines = (char *) &nlines;

	if (use3d)
	{
		type = 32768 + 5;	
	}
	else
	{
		type =  5;	
	}

	if (flipbytes)
	{
		flip_endian_int32((char *) &type);
		flip_endian_int32((char *) &nlines);
	}

	//how many points in the entire multiline()
	for (t=0;t<numb_lines; t++)
	{
		total_points += lines[t]->npoints;
	}
	

	if (use3d)
	{
		total_size = 9+ 9*numb_lines + 24*total_points;	
	}
	else
	{
		total_size = 9 + 9*numb_lines + 16*total_points;		
	}


	result = palloc (total_size);



	result[0] = byte_order;	
	result[1] = c_type[0];		//type
	result[2] = c_type[1];
	result[3] = c_type[2];
	result[4] = c_type[3];

	result[5] = c_nlines[0];	//num linestrings
	result[6] = c_nlines[1];
	result[7] = c_nlines[2];
	result[8] = c_nlines[3];


	
	new_spot = result + 9;//where to put the next linestring

		for (t=0;t<numb_lines;t++)  //for each linestring
		{
			wkb_line( lines[t], &size_part, flipbytes, byte_order, use3d, new_spot);
			new_spot += size_part; //move to next slot
		}
	return result;
}


//mem should be zero or hold enough space
char	*wkb_polygon(POLYGON3D	*poly,int32 *size, bool flipbytes, char byte_order,bool use3d, char *mem)
{
	int	t, u,total_points =0;
	uint32	type ; 	
	char		*c_type = (char *) &type;
	uint32	numRings,npoints;
	char		*c_numRings = (char *) &numRings;
	char		*c_npoints = (char *) &npoints;
	int		offset;
	POINT3D	*pts;
	int		point_offset;
	char 	*result;


	//where the polygon's points are (double aligned)

	pts = (POINT3D *) ( (char *)&(poly->npoints[poly->nrings] )  );
	pts = (POINT3D *) MAXALIGN(pts);

	

	numRings = poly->nrings;

	//total points in all the rings
	for (t=0; t< poly->nrings; t++)
	{
		total_points += poly->npoints[t];
	}

	if (use3d)
	{
		*size = 9 + 4*poly->nrings + 24*total_points;
		type = 32768 + 3;
	}
	else
	{
		*size = 9 + 4*poly->nrings + 16*total_points;
		type = 3;
	}

	if (mem == NULL)
		result = palloc (*size);
	else
		result = mem;

	
	if (flipbytes)
	{
		flip_endian_int32((char *) &type);
		flip_endian_int32((char *) &numRings);
	}

	

	result[0] = byte_order;	
	result[1] = c_type[0];		//type
	result[2] = c_type[1];
	result[3] = c_type[2];
	result[4] = c_type[3];

	result[5] = c_numRings[0];	//npoints
	result[6] = c_numRings[1];
	result[7] = c_numRings[2];
	result[8] = c_numRings[3];

	
	if (use3d)
	{
		offset = 9; //where the actual linearring structure is going
		point_offset = 0;
		for (t=0;t<poly->nrings; t++) //for each ring
		{
			npoints = poly->npoints[t];
			if (flipbytes)
				flip_endian_int32((char *) &npoints);
			result[offset]   = c_npoints[0];  //num points in this ring
			result[offset+1] = c_npoints[1];
			result[offset+2] = c_npoints[2];
			result[offset+3] = c_npoints[3];

			memcpy(&result[offset+4], &pts[point_offset], 24*poly->npoints[t]);
			if (flipbytes)
			{
				for (u=0;u<poly->npoints[t];u++)  //for each point
				{
					flip_endian_double((char *) &result[offset+4+u*24]   );
					flip_endian_double((char *) &result[offset+4+u*24+8] );
					flip_endian_double((char *) &result[offset+4+u*24]+16);
				}
			}
			point_offset += poly->npoints[t];     // reset offsets
			offset +=  4 +  poly->npoints[t]* 24;
		}
	}
	else
	{
		offset = 9; // where the 1st linearring goes
		point_offset = 0;
		for (t=0;t<poly->nrings; t++)
		{
			npoints = poly->npoints[t];
			if (flipbytes)
				flip_endian_int32((char *) &npoints);
			result[offset]   = c_npoints[0];
			result[offset+1] = c_npoints[1];
			result[offset+2] = c_npoints[2];
			result[offset+3] = c_npoints[3];
				
				for (u=0;u<poly->npoints[t];u++)
				{
					memcpy(&result[offset+4 + u*16], &pts[point_offset+u], 16);
					if (flipbytes)
					{
						flip_endian_double((char *) &result[offset+4+u*16]   );
						flip_endian_double((char *) &result[offset+4+u*16+8] );
					}
				}
			
			point_offset += poly->npoints[t];
			offset +=  4 +  poly->npoints[t]* 16;
		}

	}
	return result;
}


//expects a list of pointer to polygon3d
char	*wkb_multipolygon(POLYGON3D	**polys,int numb_polys,int32 *size, bool flipbytes, char byte_order,bool use3d)
{

	int	t,u;
	int	total_points = 0;
	int	total_size=0;
	int	size_part = 0;
	int	total_rings = 0;
	char	*result,*new_spot;
	uint32	type;

	char		*c_type = (char *) &type;

	uint32	npolys = numb_polys;
	char		*c_npolys = (char *) &npolys;

	if (use3d)
	{
		type = 32768 + 6;	
	}
	else
	{
		type =  6;	
	}


	if (flipbytes)
	{
		flip_endian_int32( (char *) &type);
		flip_endian_int32((char *) &npolys);
	}




	//find total # rings and total points
	for (t=0;t<numb_polys; t++)
	{
		total_rings += polys[t]->nrings;
		for (u=0;u<polys[t]->nrings; u++)
		{
			total_points += polys[t]->npoints[u];
		}
	}
	

	if (use3d)
	{
		total_size =9 +  9*numb_polys + 24*total_points + 4*total_rings;	
	}
	else
	{
		total_size = 9 + 9*numb_polys + 16*total_points + 4*total_rings;		
	}

	result = palloc (total_size);

	result[0] = byte_order;	
	result[1] = c_type[0];		//type
	result[2] = c_type[1];
	result[3] = c_type[2];
	result[4] = c_type[3];

	result[5] = c_npolys[0];	//number of polygons
	result[6] = c_npolys[1];
	result[7] = c_npolys[2];
	result[8] = c_npolys[3];



	new_spot = result+9;  //where 1st polygon goes

		for (t=0;t<numb_polys;t++)  //for each polygon
		{
			wkb_polygon( polys[t], &size_part, flipbytes, byte_order, use3d, new_spot);
			new_spot +=  size_part; //more to next slot
		}
	return result;

}


//passes work off to the correct function 
// then makes a simple postgres type - the first 4 bytes is an int32 with the size of the structure
// and the rest is the wkb info.
// 
// 3d-ness is determined by looking at the is3d flag in the GEOMETRY structure

char	*to_wkb(GEOMETRY *geom, bool flip_endian)
{
	char	*result, *sub_result;
	int 	size;
	
	if (geom->type == COLLECTIONTYPE)
	{
		sub_result=	to_wkb_collection(geom,  flip_endian, &size);
		size += 4; //size info
		result = palloc (size );
		memcpy(result+4, sub_result, size-4);
		memcpy(result, &size, 4);
		pfree(sub_result);
		return result;
	}
	else
	{
		sub_result=	to_wkb_sub(geom,  flip_endian, &size);
		size += 4; //size info
		result = palloc (size );
		memcpy(result+4, sub_result, size-4);
		memcpy(result, &size, 4);
		pfree(sub_result);
		return result;	
	}

}

//make a wkb chunk out of a geometrycollection
char *to_wkb_collection(GEOMETRY *geom, bool flip_endian, int32 *end_size)
{
	char	*result, **sub_result;
	int	sub_size;
	int	*sizes;
	int	t;
	int32		*offsets1;
	
	int32		type;
	int		total_size =0;
	int		offset;
	char		byte_order;
	int		coll_type=7;
	char		*c_type = (char *)&coll_type;
	int		nobj = geom->nobjs;
	char		*c_nobj = (char *) &nobj;

//printf("making a wkb of a collections\n");


	//determing byte order and if numbers need endian change
	if (BYTE_ORDER == BIG_ENDIAN)
	{
		if (flip_endian)
		{
			byte_order = 1;
		}
		else
		{
			byte_order = 0;
		}
	}
	else
	{
		if (flip_endian)
		{
			byte_order = 0;
		}
		else
		{
			byte_order = 1;
		}
	}

	//for sub part of the geometry structure
	offsets1 = (int32 *) ( ((char *) &(geom->objType[0] ))+ sizeof(int32) * geom->nobjs ) ;


	// we make a list of smaller wkb chunks in sub_result[]
	// and that wkb chunk's size in sizes[]


	sub_result = palloc( sizeof(char *) * geom->nobjs);
	sizes = palloc( sizeof(int) * geom->nobjs);


	for (t=0; t<=geom->nobjs; t++)	//for each part of the collections, do the work in another function
	{
		type = geom->objType[t];		

		if (type == POINTTYPE)
		{
			sub_result[t] = (  wkb_point((POINT3D *) ((char *) geom +offsets1[t] ) , 
					&sub_size, flip_endian, byte_order, geom->is3d));
			sizes[t] = sub_size;
			total_size += sub_size;
		}

		if (type == LINETYPE)
		{

			sub_result[t] = (  wkb_line((LINE3D *) ((char *) geom +offsets1[t] ) , 
				 &sub_size, flip_endian, byte_order, geom->is3d, NULL));
			sizes[t] = sub_size;
			total_size += sub_size;
		}
		if (type == POLYGONTYPE)
		{

			sub_result[t] = (  wkb_polygon((POLYGON3D *) ((char *) geom +offsets1[t] ) , 
				 &sub_size, flip_endian, byte_order, geom->is3d, NULL));
			sizes[t] = sub_size;
			total_size += sub_size;
		}
	}

	result = palloc( total_size +9); // need geometrycollection header

	if (flip_endian)  
	{
		flip_endian_int32((char *) &coll_type);
		flip_endian_int32((char *) &nobj);
	}

	//could use a memcpy, but...

	result[0] = byte_order;	
	result[1] = c_type[0];		//type
	result[2] = c_type[1];
	result[3] = c_type[2];
	result[4] = c_type[3];

	result[5] = c_nobj[0];	//number of part to collection
	result[6] = c_nobj[1];
	result[7] = c_nobj[2];
	result[8] = c_nobj[3];
	

	offset =9;  //start after the geometrycollection header


	for (t=0;t<geom->nobjs; t++)
	{
		memcpy(result+ offset, sub_result[t], sizes[t]);
		pfree(sub_result[t]);
		offset += sizes[t];
	}

		//free temp structures
	pfree( sub_result);
	pfree( sizes);

		//total size of the wkb 
	*end_size = total_size;



	//decode_wkb(result, &junk);
	return result;
}


//convert geometry to WKB, flipping endian if appropriate
//
//  This handles POINT, LINESTRING, POLYGON, MULTIPOINT,MULTILINESTRING, MULTIPOLYGON
//  
// GEOMETRYCOLLECTION is not handled by this function see to_wkb_collection()
//
// flip_endian - request an endian swap (relative to current machine)
// wkb_size    - return size of wkb

char	*to_wkb_sub(GEOMETRY *geom, bool flip_endian, int32 *wkb_size)
{

	char	byte_order;
	char	*result;
	int		t;
	

	int32		*offsets1;
	LINE3D	**linelist;
	POLYGON3D	**polylist;


	//determine WKB byte order flag
	if (BYTE_ORDER == BIG_ENDIAN)	//server is running on a big endian machine
	{
		if (flip_endian)
		{
			byte_order = 1;  //as per WKB specification
		}
		else
		{
			byte_order = 0;
		}
	}
	else
	{
		if (flip_endian)
		{
			byte_order = 0;
		}
		else
		{
			byte_order = 1;
		}
	}

	// for searching in the geom struct	
	offsets1 = (int32 *) ( ((char *) &(geom->objType[0] ))+ sizeof(int32) * geom->nobjs ) ;


		// for each of the possible geometry types, we pass it off to a worker function

	if (geom->type == POINTTYPE)
	{

		result = (  wkb_point((POINT3D *) ((char *) geom +offsets1[0] ) , 
					wkb_size, flip_endian, byte_order, geom->is3d));
	}

	if (geom->type == MULTIPOINTTYPE)
	{

		result = (  wkb_multipoint((POINT3D *) ((char *) geom +offsets1[0] ) , 
				geom->nobjs,	wkb_size, flip_endian, byte_order, geom->is3d));
	}

	if (geom->type == LINETYPE)
	{

		result = (  wkb_line((LINE3D *) ((char *) geom +offsets1[0] ) , 
				 wkb_size, flip_endian, byte_order, geom->is3d, NULL));
	}


	if (geom->type == MULTILINETYPE)
	{
		//make a list of lines
		linelist = palloc( sizeof(LINE3D *) * geom->nobjs);
		for (t=0;t<geom->nobjs; t++)
		{
			linelist[t] = (LINE3D *) ((char *) geom +offsets1[t] ) ;
		}
		result = (  wkb_multiline( linelist,
				 wkb_size, geom->nobjs, flip_endian, byte_order, geom->is3d));
	}

	if (geom->type == POLYGONTYPE)
	{

		result = (  wkb_polygon((POLYGON3D *) ((char *) geom +offsets1[0] ) , 
				 wkb_size, flip_endian, byte_order, geom->is3d, NULL));
	}


	if (geom->type == MULTIPOLYGONTYPE)
	{
		//make a list of polygons
		polylist = palloc( sizeof(POLYGON3D *) * geom->nobjs);
		for (t=0;t<geom->nobjs; t++)
		{
			polylist[t] = (POLYGON3D *) ((char *) geom +offsets1[t] ) ;
		}
		result = (  wkb_multipolygon( polylist,geom->nobjs,
				 wkb_size,  flip_endian, byte_order, geom->is3d));
	}

	//decode_wkb(result, &junk);

	return (result);
}


//convert binary geometry into OGIS well know binary format with XDR (big endian) formating
// see http://www.opengis.org/techno/specs/99-049.rtf page 3-24 for specification
//
// 3d geometries are encode as in OGR by adding 32768 to the type.  Points are then 24bytes (X,Y,Z)
// instead of 16 bytes (X,Y)

PG_FUNCTION_INFO_V1(wkb_XDR);
Datum wkb_XDR(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));

	if (BYTE_ORDER == BIG_ENDIAN)
		PG_RETURN_POINTER(to_wkb(geom, FALSE));
	else
		PG_RETURN_POINTER(to_wkb(geom, TRUE));
}



//convert binary geometry into OGIS well know binary format with NDR (little endian) formating
// see http://www.opengis.org/techno/specs/99-049.rtf page 3-24 for specification
//
// 3d geometries are encode as in OGR by adding 32768 to the type.  Points are then 24bytes (X,Y,Z)
// instead of 16 bytes (X,Y)

PG_FUNCTION_INFO_V1(wkb_NDR);
Datum wkb_NDR(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));

	if (BYTE_ORDER == LITTLE_ENDIAN)
		PG_RETURN_POINTER(to_wkb(geom, FALSE));
	else
		PG_RETURN_POINTER(to_wkb(geom, TRUE));

}

//  DO NOT USE THESE decoding function (except for debugging purposes)
//    The code is NOT maintained and is thrown together


// unsupported debugging function.
// given a wkb input thats a geometrycollection, returns its size and prints out
// its contents
//
//  Its really messy - dont even think about using this for anything
// 
// you shouldnt call this function; just call decode_wkb() and it will
// call this function
//
void decode_wkb_collection(char *wkb,int	*size)
{
	int	offset =0;
	bool	flipbytes;
	int	total_size=0,sub_size;
	int	numb_sub,t;
	bool	first_one = TRUE;
	
		
	if (wkb[0] ==0 )  //big endian
	{
		if (BYTE_ORDER == LITTLE_ENDIAN)
			flipbytes= 1;
		else
			flipbytes= 0;
	}
	else //little
	{
		if (BYTE_ORDER == LITTLE_ENDIAN)
			flipbytes= 0;
		else
			flipbytes= 1;
	}

	memcpy(&numb_sub, wkb+5,4);
	if (flipbytes)
		flip_endian_int32( (char *) & numb_sub) ;

	printf("GEOMETRYCOLLECTION(\n");
	offset = 9;
	for (t=0;t<numb_sub;t++)
	{
			if (first_one)
			{
				first_one = FALSE;
			}
			else
			{
				printf(",");
			}

		printf("	");
		decode_wkb( wkb + offset, &sub_size);
		total_size += sub_size;
		offset += sub_size;
	}
	*size = total_size +9 ;
	printf(")\n");
	return;
}


// unsupported debugging function.
// given a wkb input, returns its size and prints out
// its contents
//
//  Its really messy - dont even think about using this for anything
//
//dump wkb to screen (for debugging)
// assume endian is constant though out structure
void decode_wkb(char *wkb, int *size)
{

	bool	flipbytes;
	uint32	type;
	uint32	n1,n2,n3,t,u,v;
	bool	is3d;
	bool first_one,first_one2,first_one3;

	int	offset,offset1;
	double	x,y,z;
	int	total_points;



	//printf("decoding wkb\n");
	

	if (wkb[0] ==0 )  //big endian
	{
		if (BYTE_ORDER == LITTLE_ENDIAN)
			flipbytes= 1;
		else
			flipbytes= 0;
	}
	else //little
	{
		if (BYTE_ORDER == LITTLE_ENDIAN)
			flipbytes= 0;
		else
			flipbytes= 1;
	}
	
	//printf("	+ flipbytes = %i\n", flipbytes);

	//printf("info about wkb:\n");
	memcpy(&type, wkb+1,4);
	if (flipbytes)
		flip_endian_int32( (char *) & type) ;
	
	is3d = 0;
 
	if (type > 1000 )
	{
		is3d = 1;
		type = type - 32768;
	}
	//printf("	Type = %i (is3d = %i)\n", type, is3d);
	if (type == 1)  //POINT()
	{
		printf("POINT(");
		if (is3d)
		{
			memcpy(&x, &wkb[5], 8);
			memcpy(&y, &wkb[5+8], 8);
			memcpy(&z, &wkb[5+16], 8);
			if (flipbytes)
			{
				flip_endian_double( (char *) & x) ;
				flip_endian_double( (char *) & y) ;
				flip_endian_double( (char *) & z) ;
			}
			printf("%g %g %g)",x,y,z );
		}
		else
		{
			memcpy(&x, &wkb[5], 8);
			memcpy(&y, &wkb[5+8], 8);
			if (flipbytes)
			{
				flip_endian_double( (char *) & x) ;
				flip_endian_double( (char *) & y) ;
			}
			printf("%g %g)", x,y);
		}
		printf("\n");
		if (is3d)	
			*size = 29;
		else
			*size = 21;
		return;
		
	}
	if (type == 2)
	{
		printf("LINESTRING(");
		memcpy(&n1, &wkb[5],4);
		if (flipbytes)
			flip_endian_int32( (char *) & n1) ;
	//	printf("  --- has %i sub points\n",n1);
		first_one = TRUE;
		for (t=0; t<n1; t++)
		{
			if (first_one)
			{
				first_one = FALSE;
			}
			else
			{
				printf(",");
			}
			if (is3d)
			{
				memcpy(&x, &wkb[9+t*24],8);
				memcpy(&y, &wkb[9+t*24+8],8);
				memcpy(&z, &wkb[9+t*24+16],8);
				if (flipbytes)
				{
					flip_endian_double( (char *) & x) ;
					flip_endian_double( (char *) & y) ;
					flip_endian_double( (char *) & z) ;
				}
				printf("%g %g %g",x,y,z);
			}
			else
			{
				memcpy(&x, &wkb[9+t*16],8);
				memcpy(&y, &wkb[9+t*16+8],8);
				if (flipbytes)
				{
					flip_endian_double( (char *) & x) ;
					flip_endian_double( (char *) & y) ;
				}
				printf("%g %g",x,y);
			}
		}


		printf(")\n");	
		if (is3d)	
			*size = 9 + n1*24;
		else
			*size = 9 + n1*16;
		return;
	}
	if (type == 3)
	{
		*size = 9;
		printf("POLYGON(");
		memcpy(&n1, &wkb[5],4);
		if (flipbytes)
			flip_endian_int32( (char *) & n1) ;
		//printf("  --- has %i rings\n",n1);
		*size += 4*n1;
		offset= 9;
		first_one = TRUE;
		for (u=0; u<n1; u++)
		{
			memcpy(&n2, &wkb[offset],4);
			if (flipbytes)
				flip_endian_int32( (char *) & n2) ;
		//	printf(" ring %i: has %i points\n",u,n2);
			
			
			if (first_one)
			{
				first_one = FALSE;
			}
			else
			{
			  	printf(",");
			}
			printf("(");

			first_one2 = TRUE;
			for (v=0; v< n2; v++)
			{
			if (first_one2)
			{
				first_one2 = FALSE;
			}
			else
			{
			  	printf(",");
			}
				if (is3d)
				{
					memcpy(&x, &wkb[offset + 4+ v*24],8);
					memcpy(&y, &wkb[offset + 4+ v*24 + 8],8);
					memcpy(&z, &wkb[offset + 4+ v*24 + 16],8);
					if (flipbytes)
					{
						flip_endian_double( (char *) & x) ;
						flip_endian_double( (char *) & y) ;
						flip_endian_double( (char *) & z) ;
					}
					printf("%g %g %g",x,y,z);

				}
				else
				{
					memcpy(&x, &wkb[offset +4 +v*16],8);
					memcpy(&y, &wkb[offset +4 +v*16 + 8],8);
					if (flipbytes)
					{
						flip_endian_double( (char *) & x) ;
						flip_endian_double( (char *) & y) ;
					}
					printf("%g %g",x,y);
				
				}
			}
			if (is3d)
			{
				offset +=4 +24*n2;
				*size += n2*24;
			}
			else
			{
				offset += 4+ 16*n2;
				*size += n2*16;
			}
			printf(")");
		}

		printf(")\n");
		
		return;

	}
	if (type == 4)
	{
		printf("MULTIPOINT(");
		memcpy(&n1,&wkb[5],4);
		if (flipbytes)
			flip_endian_int32( (char *) & n1) ;
	//	printf(" -- has %i points\n",n1);
		if (is3d)	
			*size = 9 + n1*29;
		else
			*size = 9 + n1*21;
		first_one = TRUE;
		for (t=0; t<n1; t++)
		{
			if (first_one)
			{
				first_one= FALSE;
			}	
			else
			{
				printf(",");
			}
			if (is3d)
			{
				memcpy(&x, &wkb[9+t*29+5],8);
				memcpy(&y, &wkb[9+t*29+8+5],8);
				memcpy(&z, &wkb[9+t*29+16+5],8);
				if (flipbytes)
				{
					flip_endian_double( (char *) & x) ;
					flip_endian_double( (char *) & y) ;
					flip_endian_double( (char *) & z) ;
				}
				printf("%g %g %g",x,y,z);
			}
			else
			{
				memcpy(&x, &wkb[9+t*21+5],8);
				memcpy(&y, &wkb[9+t*21+8+5],8);
				if (flipbytes)
				{
					flip_endian_double( (char *) & x) ;
					flip_endian_double( (char *) & y) ;
				}
				printf("%g %g",x,y);
			}
		}
		printf (")\n");
		return;
	}
	if (type == 5)
	{
		*size = 9;
		printf("MULTILINESTRING(");
		memcpy(&n2,&wkb[5],4);
		if (flipbytes)
			flip_endian_int32( (char *) & n2) ;
	//	printf(" -- has %i sub-lines\n",n2);
		*size += 9 *n2;
		offset =9;
		first_one2 = TRUE;
		for (u=0; u<n2; u++)
		{
			if (first_one2)
			{
				first_one2= FALSE;
			}	
			else
			{
				printf(",");
			}
		printf("(");
		memcpy(&n1, &wkb[5 +offset],4);
		if (flipbytes)
			flip_endian_int32( (char *) & n1) ;
	//	printf("  --- has %i sub points\n",n1);
		first_one = TRUE;
		for (t=0; t<n1; t++)
		{
			if (first_one)
			{
				first_one= FALSE;
			}	
			else
			{
				printf(",");
			}

			if (is3d)
			{
				memcpy(&x, &wkb[offset+9+t*24],8);
				memcpy(&y, &wkb[offset+9+t*24+8],8);
				memcpy(&z, &wkb[offset+9+t*24+16],8);
				if (flipbytes)
				{
					flip_endian_double( (char *) & x) ;
					flip_endian_double( (char *) & y) ;
					flip_endian_double( (char *) & z) ;
				}
				printf("%g %g %g",x,y,z);
			}
			else
			{
				memcpy(&x, &wkb[offset+9+t*16],8);
				memcpy(&y, &wkb[offset+9+t*16+8],8);
				if (flipbytes)
				{
					flip_endian_double( (char *) & x) ;
					flip_endian_double( (char *) & y) ;
				}
				printf("%g %g",x,y);
			}
		}
		printf(")");
		if (is3d)
		{
			*size += (24*n1);
			offset += 9 + (24*n1);
		}
		else
		{
			*size += (16*n1);
			offset += 9 + (16*n1);
		}
		}
		printf(")\n");
		return;

	}
	if (type == 6)
	{
		*size = 9;
		printf("MULTIPOLYGON(");
		memcpy(&n3,&wkb[5],4);
		if (flipbytes)
			flip_endian_int32( (char *) & n3) ;
		//printf(" -- has %i sub-poly\n",n3);
		*size += 9*n3;
		offset1 =9;//where polygon starts
		first_one3= TRUE;
	for (t=0;t<n3; t++)  //for each polygon
	{
			if (first_one3)
			{
				first_one3= FALSE;
			}	
			else
			{
				printf(",");
			}
		printf("(");
		//printf("polygon #%i\n",t);
		total_points = 0;
		memcpy(&n1,&wkb[offset1+5],4); //# rings
		*size += 4*n1;
		if (flipbytes)
			flip_endian_int32( (char *) & n1) ;
		//printf("This polygon has %i rings\n",n1);
		offset = offset1+9; //where linear rings are
		first_one = TRUE;
		for (u=0; u<n1; u++) //for each ring
		{
			if (first_one)
			{
				first_one= FALSE;
			}	
			else
			{
				printf(",");
			}
			printf("(");
			memcpy(&n2, &wkb[offset],4);
			if (flipbytes)
				flip_endian_int32( (char *) & n2) ; //pts in linear ring
		//	printf(" ring %i: has %i points\n",u,n2);
			total_points += n2;
			first_one2 = TRUE;
			for (v=0; v< n2; v++)	//for each point
			{
			if (first_one2)
			{
				first_one2= FALSE;
			}	
			else
			{
				printf(",");
			}
				if (is3d)
				{
					memcpy(&x, &wkb[offset + 4+ v*24],8);
					memcpy(&y, &wkb[offset + 4+ v*24 + 8],8);
					memcpy(&z, &wkb[offset + 4+ v*24 + 16],8);
					if (flipbytes)
					{
						flip_endian_double( (char *) & x) ;
						flip_endian_double( (char *) & y) ;
						flip_endian_double( (char *) & z) ;
					}
					printf("%g %g %g",x,y,z);
				}
				else
				{
					memcpy(&x, &wkb[offset +4 +v*16],8);
					memcpy(&y, &wkb[offset +4 +v*16 + 8],8);
					if (flipbytes)
					{
						flip_endian_double( (char *) & x) ;
						flip_endian_double( (char *) & y) ;
					}
					printf("%g %g",x,y);
				}
			}
			if (is3d)
			{
				*size += 24*n2;
				offset += 4+ 24*n2;
			}
			else
			{
				*size += 16*n2;
				offset += 4+ 16*n2;			
			}
			printf(")");
		}
		printf(")");
		if (is3d)
			offset1 +=9 +24*total_points +4*n1;
		else
			offset1 += 9+ 16*total_points +4*n1;
	}
	printf(")\n");
	return;
	}
	if (type == 7)
	{
		return	decode_wkb_collection(wkb, size);
	}
}


// set the is3d flag so the system think the geometry is 2d or 3d

PG_FUNCTION_INFO_V1(force_2d);
Datum force_2d(PG_FUNCTION_ARGS)
{
	GEOMETRY		      *geom = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));

	geom->is3d = FALSE;
	PG_RETURN_POINTER(geom);
}


PG_FUNCTION_INFO_V1(force_3d);
Datum force_3d(PG_FUNCTION_ARGS)
{
		GEOMETRY		      *geom = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));

	geom->is3d = TRUE;
	PG_RETURN_POINTER(geom);

}


//NULL safe bbox union
// combine_bbox(BOX3D, geometry) => union BOX3D and geometry->bvol
// For finding the extent of a set of rows using the extent() aggregate

PG_FUNCTION_INFO_V1(combine_bbox);
Datum combine_bbox(PG_FUNCTION_ARGS)
{
		 Pointer        	 box3d_ptr = PG_GETARG_POINTER(0);
		 Pointer        	 geom_ptr =  PG_GETARG_POINTER(1);
		 BOX3D			*a,*b;
		 GEOMETRY			*geom;
		BOX3D				*box;

	if  ( (box3d_ptr == NULL) && (geom_ptr == NULL) )
	{
		PG_RETURN_NULL(); // combine_bbox(null,null) => null
	}

	if (box3d_ptr == NULL)
	{
		geom = (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
		PG_RETURN_POINTER(&(geom->bvol));  // combine_bbox(null, geometry) => geometry->bvol
	}

	if (geom_ptr == NULL)
	{
		PG_RETURN_POINTER( PG_GETARG_DATUM(0) ); // combine_bbox(BOX3D, null) => BOX3D
	}

	//combine_bbox(BOX3D, geometry) => union(BOX3D, geometry->bvol)

	box = (BOX3D *) palloc(sizeof(BOX3D));

	a = (BOX3D *) PG_GETARG_DATUM(0) ;
	b= &(( (GEOMETRY *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1)) )->bvol);

	box->LLB.x = b->LLB.x;
	box->LLB.y = b->LLB.y;
	box->LLB.z = b->LLB.z;

	box->URT.x = b->URT.x;
	box->URT.y = b->URT.y;
	box->URT.z = b->URT.z;

	if (a->LLB.x < b->LLB.x)
		box->LLB.x = a->LLB.x;
	if (a->LLB.y < b->LLB.y)
		box->LLB.y = a->LLB.y;
	if (a->LLB.z < b->LLB.z)
		box->LLB.z = a->LLB.z;

	if (a->URT.x > b->URT.x)
		box->URT.x = a->URT.x;
	if (a->URT.y > b->URT.y)
		box->URT.y = a->URT.y;
	if (a->URT.z > b->URT.z)
		box->URT.z = a->URT.z;

	PG_RETURN_POINTER(  box  );
}

